# Copyright 2020-present Tae Hwan Jung
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

_KB = 1024
"""The size of a Kilobyte in bytes"""

_MB = 1024 * _KB
"""The size of a Megabyte in bytes"""

import os
import io
import h5py
import torch
from collections import OrderedDict

from matorage.model.manager import Manager


class ModelManager(Manager):

    """
    Model Manager Pytorch classes. This class overrides ``Manager``.

    Note:
        Unlike Dataset, model weight is loaded entirely into cpu memory.
        Therefore, the `HDF5_CORE` driver using the memory option is default setting.

    Args:
        config (:obj:`matorage.ModelConfig`, **require**):
            A ModelConfig instance object
        num_worker_threads (:obj:`int`, optional, defaults to `4`):
            Number of backend storage worker to upload or download.
        multipart_upload_size (:obj:`integer`, optional, defaults to `5 * 1024 * 1024`):
            size of the incompletely uploaded object.
            You can sync files faster with `multipart upload in MinIO. <https://github.com/minio/minio-py/blob/master/minio/api.py#L1795>`_
            This is because MinIO clients use multi-threading, which improves IO speed more
            efficiently regardless of Python's Global Interpreter Lock(GIL).

    Examples::

        from matorage import ModelConfig
        from matorage.torch import ModelManager

        model_config = ModelConfig(
            endpoint='127.0.0.1:9000',
            access_key='minio',
            secret_key='miniosecretkey',
            model_name='testmodel',
            additional={
                "version" : "1.0.1"
            }
        )

        model_manager = ModelManager(config=model_config)

        import torch.nn as nn
        class Model(nn.Module):
            def __init__(self):
                super(Model, self).__init__()
                self.net1 = nn.Sequential(
                    nn.Linear(5, 10),
                    nn.ReLU(),
                )
                self.net2 = nn.Sequential(
                    nn.Linear(10, 5),
                    nn.ReLU(),
                )
            def forward(self, x):
                x = self.net1(x)
                x = self.net2(x)
                return x

        model = Model()
        model_manager.save(model, step=100)


    """

    def __init__(self, config, num_worker_threads=4, multipart_upload_size=5 * _MB):
        super(ModelManager, self).__init__(
            config, num_worker_threads, multipart_upload_size
        )

    def _save_model(self, model_folder, model):
        for name, weight in model.state_dict().items():
            self._save_layer(model_folder, name, weight.cpu().numpy())

    def _load_model(self, model_folder, layers, model):
        weight = OrderedDict()

        if isinstance(model, str):
            keys = [model]
        else:
            keys = list(model.state_dict().keys())

        for layer in layers:
            name = os.path.basename(layer.object_name)
            if name in keys:
                layer_image = self._client.get_object(
                    bucket_name=self.config.bucket_name,
                    object_name=f"{model_folder}/{name}",
                ).read()

                layer_image = h5py.File(io.BytesIO(layer_image), "r")
                weight[name] = torch.from_numpy(layer_image[self.type][:])

        if isinstance(model, str):
            return weight
        else:
            model.load_state_dict(weight)

    def save(self, model, **kwargs):
        """
        save weight of model

        Args:
            model (:obj:`torch.nn.Module`, **require**):
                Pytorch model (``torch.nn.Module`` type)
            kwargs (:obj:`**kwargs`, optional):
                metadata about step or epoch for model.

        Examples::

            >>> model_manager.save(model, step=0)
            >>> model_manager.save(model, epoch=0)
            >>> model_manager.save(model, epoch=0, step=0)

        """
        super(ModelManager, self).save(model, **kwargs)

    def load(self, model, **kwargs):
        """
        load weight of model

        Args:
            model (:obj:`torch.nn.Module` or `string`, **require**):
                Pytorch model(``torch.nn.Module`` type) or layer name(string type).
            kwargs (:obj:`**kwargs`, optional):
                metadata about step or epoch for model.

        Returns:
            :obj:`None or OrderedDict`: If ``model`` is pytorch model, weight is loaded into the model and return None.
            however, If it is a string type with the name of the layer, it returns the weight of the OrderedDict type.

        Examples::

            # Load entire model
            >>> model = Model()
            >>> model_manager.load(model, step=0)
            >>> model
            Model(
              (net1): Sequential(
                (0): Linear(in_features=5, out_features=10, bias=True)
                (1): ReLU()
              )
              (net2): Sequential(
                (0): Linear(in_features=10, out_features=5, bias=True)
                (1): ReLU()
              )
            )

            # Load sub-layer model
            >>> class SubModel(nn.Module):
            >>>   def __init__(self):
            >>>     super(SubModel, self).__init__()
            >>>     self.net1 = nn.Sequential(
            >>>         nn.Linear(5, 10),
            >>>         nn.ReLU(),
            >>>     )

            >>> submodel = SubModel()
            >>> model_manager.load(submodel, step=0)
            >>> model
            Model(
              (net1): Sequential(
                (0): Linear(in_features=5, out_features=10, bias=True)
                (1): ReLU()
              )
            )

            # Load from layer name
            >>> model_manager.load('net1.0.weight', step=0)
            OrderedDict([('net1.0.weight',
            tensor([[ 0.2679, -0.2147, -0.1927, -0.3263,  0.0930],
                [ 0.0144,  0.2935,  0.3614, -0.0493, -0.3772],
                [ 0.4101, -0.1864,  0.1076, -0.3900,  0.3613],
                [-0.2831,  0.3692,  0.3367,  0.2491, -0.2971],
                [-0.3019,  0.1682, -0.3951,  0.1528,  0.1778],
                [-0.1593,  0.3315, -0.2286,  0.1294,  0.2087],
                [-0.3394, -0.2706,  0.1515,  0.0357, -0.4252],
                [ 0.2555, -0.4435, -0.3353,  0.2096, -0.3741],
                [ 0.3950, -0.2630, -0.1730,  0.1393,  0.3678],
                [ 0.3065, -0.0095,  0.0988,  0.4294,  0.3338]]))])

        """
        return super(ModelManager, self).load(model, **kwargs)



# Copyright 2020-present Tae Hwan Jung
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

from matorage.serialize import Serialize


class DataMetadata(Serialize):
    r""" Metadata of dataset configuration classes.
        Handles a few parameters configuration for only dataset.
        This class is stored together in the MinIO in json format.

        Note:
            This class is recommended not to be used in the code of the user.
            **`DataConfig` must be mapped with only one `DataMetadata`.**

        About indexer's work
            we manages index to dict type as total data length as key and file names as value.
            Examples:
                "indexer": {
                    "15960": "tmpjopm9_qr118171adec9d410d.h5",
                    "31920": "tmpnzdga4zfc952613ed2bf4fad.h5",
                    "47880": "tmpy3fwohhx178eb7c1d17d4a78.h5",
                    "60000": "tmpidco3y8u794f8adde1e5479a.h5"
                }
    """

    def __init__(self, **kwargs):
        self.endpoint = kwargs["endpoint"]
        self.dataset_name = kwargs.pop("dataset_name", None)
        self.additional = kwargs.pop("additional", {})

        self.attributes = kwargs.pop("attributes", None)
        self.compressor = kwargs.pop("compressor", {"complevel": 0, "complib": "zlib"})

        self.indexer = {}
        self.filetype = []

    def to_dict(self):
        """
        Serializes this instance to a Python dictionary.

        Returns:
            :obj:`Dict[str, any]`: Dictionary of all the attributes that make up this configuration instance,
        """
        output = copy.deepcopy(self.__dict__)
        if hasattr(self.__class__, "attributes") or "attributes" in output:
            output["attributes"] = [
                _attribute.to_dict() for _attribute in self.attributes
            ]
        if (
            hasattr(self.__class__, "flatten_attributes")
            or "flatten_attributes" in output
        ):
            output["flatten_attributes"] = [
                _attribute.to_dict() for _attribute in self.flatten_attributes
            ]
        return output



# Original Code
# https://github.com/huggingface/transformers/blob/master/src/transformers/configuration_utils.py
# See https://github.com/graykode/matorage/blob/0.1.0/NOTICE
# modified by TaeHwan Jung(@graykode)

# Copyright 2020-present Tae Hwan Jung
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import json


class Serialize(object):
    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "{} {}".format(self.__class__.__name__, self.to_json_string())

    def to_dict(self):
        """
        Serializes this instance to a Python dictionary.

        Returns:
            :obj:`Dict[str, any]`: Dictionary of all the attributes that make up this configuration instance,
        """
        return copy.deepcopy(self.__dict__)

    def to_json_file(self, json_file_path):
        """
        Save this instance to a json file.

        Args:
            json_file_path (:obj:`string`):
                Path to the JSON file in which this configuration instance's parameters will be saved.
        """
        with open(json_file_path, "w", encoding="utf-8") as writer:
            writer.write(self.to_json_string())

    def to_json_string(self):
        """
        Serializes this instance to a JSON string.

        Returns:
            :obj:`string`: String containing all the attributes that make up this configuration instance in JSON format.
        """
        return json.dumps(self.to_dict(), indent=4, sort_keys=True) + "\n"

    @classmethod
    def from_json_file(cls, json_file):
        """
        Constructs a `Config` from the path to a json file of parameters.

        Args:
            json_file (:obj:`string`):
                Path to the JSON file containing the parameters.

        Returns:
            :class:`DataConfig ModelConfig`: An instance of a configuration object

        """
        config_dict = cls._dict_from_json_file(json_file)

        return cls(**config_dict)

    @classmethod
    def _dict_from_json_file(cls, json_file):
        with open(json_file, "r", encoding="utf-8") as reader:
            text = reader.read()
        return json.loads(text)



# Original Code
# https://github.com/PyTables/PyTables/blob/master/tables/tests/test_suite.py
# See https://github.com/graykode/matorage/blob/0.1.0/NOTICE
# modified by TaeHwan Jung(@graykode)

import sys
import unittest

from matorage.utils import is_torch_available, is_tf_available


def suite():
    test_modules = [
        "tests.test_datasaver",
        "tests.test_nas",
    ]
    if is_torch_available():
        test_modules.extend([
            "tests.test_torch_data",
            "tests.test_torch_model",
            "tests.test_torch_optimizer",
        ])
    if is_tf_available():
        test_modules.extend([
            "tests.test_tf_data",
            "tests.test_tf_model",
            "tests.test_tf_optimizer",
        ])

    alltests = unittest.TestSuite()
    for name in test_modules:
        # Unexpectedly, the following code doesn't seem to work anymore
        # in python 3
        # exec('from %s import suite as test_suite' % name)
        __import__(name)
        test_suite = sys.modules[name].suite
        alltests.addTest(test_suite())
    return alltests

if __name__ == '__main__':
    unittest.main(defaultTest="suite", warnings=False)



# Copyright 2020-present Tae Hwan Jung
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import io
import tensorflow as tf
import tensorflow_io as tfio

from matorage.data.data import MTRData
from matorage.utils import logger


class Dataset(MTRData):
    """
    Dataset class for Tensorflow Dataset

    This class is customized for the dataset of the PyTorch, so it is operated by the following procedure.

    1. The ``_object_file_mapper`` manages the minio object as key and the downloaded local path as value. \
    When minio object is downloaded, it is recorded in ``_object_file_maper``.
    2. We read ``_object_file_mapper`` and download only new objects that are not there.
    3. if Tensorflow v2(2.2.0>=), we use ``tfio.IODataset.from_hdf5`` and parallel ``interleave`` more fast

    Args:
        config (:obj:`matorage.DataConfig`, **require**):
            dataset configuration
        num_worker_threads (:obj:`int`, optional, defaults to `4`):
            Number of backend storage worker to upload or download.
        clear (:obj:`boolean`, optional, defaults to `True`):
            Delete all files stored on the local storage after the program finishes.
        cache_folder_path (:obj:`str`, optional, defaults to `~/.matorage`):
            Cached folder path to check which files are downloaded complete.
        index (:obj:`boolean`, optional, defaults to `False`):
            Setting for index mode.

        batch_size (:obj:`integer`, `optional`, defaults to `1`):
            how many samples per batch to load.
        shuffle (:obj:`boolean`, `optional`, defaults to `False`):
            set to True to have the data reshuffled at every epoch.
        seed (:obj:`integer`, `optional`, defaults to `0`):
            random seed used to shuffle the sampler if ``shuffle=True``.

    Examples::

        from matorage import DataConfig
        from matorage.tensorflow import Dataset

        data_config = DataConfig(
            endpoint='127.0.0.1:9000',
            access_key='minio',
            secret_key='miniosecretkey',
            dataset_name='array_test',
            attributes=[
                ('array', 'uint8', (3, 224, 224)),
            ]
        )

        dataset = Dataset(config=data_config, clear=True)

        # iterative mode
        for array in dataset.dataloader:
            print(array)

        # index mode
        print(dataset[0])

    """

    def __init__(self, config, batch_size=1, **kwargs):

        # class parameters
        self._batch_size = batch_size
        self._shuffle = kwargs.pop("shuffle", False)
        self._seed = kwargs.pop("seed", 0)
        self.index = kwargs.pop("index", False)

        super(Dataset, self).__init__(config, **kwargs)

        if not self.index:
            _dataset = tf.data.Dataset.from_tensor_slices(self.filenames)
            if self._shuffle:
                _dataset = _dataset.shuffle(len(self.filenames), seed=self._seed)
            self._dataloader = _dataset.interleave(
                self._get_item_with_download, cycle_length=len(self.filenames)
            )

    def __getitem__(self, idx):
        return self._get_item_with_inmemory(idx)

    def _get_item_with_download(self, filename):
        _tfios = []
        for _attr_name, _attr_value in self.attribute.items():
            _tfios.append(
                tfio.IODataset.from_hdf5(
                    filename,
                    dataset=f"/{_attr_name}",
                    spec=tf.as_dtype(_attr_value["type"]),
                ).map(
                    lambda x: tf.reshape(x, _attr_value["shape"])
                    if not _attr_value["shape"] == [1]
                    else x[0]
                )
            )
        _tfiodataset = tf.data.Dataset.zip(tuple(_tfios))
        if self._shuffle:
            _tfiodataset = _tfiodataset.shuffle(1000, seed=self._seed)
        _tfiodataset = _tfiodataset.batch(
            self._batch_size, drop_remainder=True
        ).prefetch(tf.data.experimental.AUTOTUNE)
        return _tfiodataset

    def _reshape_convert_tensor(self, numpy_array, attr_name):
        """
        Reshape numpy tensor and convert from numpy to torch tensor.
        In matorage dataset save in 2D (bz, N) shape to cpu L1 cache manage dataset fast.
        Therefore, this function restores the shape for the user to use.

        Returns:
            :obj:`tf.tensor`
        """
        _shape = self.attribute[attr_name]["shape"]
        numpy_array = numpy_array.reshape(_shape)
        tensor = tf.convert_to_tensor(numpy_array)
        return tensor

    @property
    def filenames(self):
        """
        Get filenames(file absolute path) in local storage

        Returns:
            :obj:`list`: filenames(file absolute path) in local storage
        """
        return list(self._object_file_mapper.values())

    @property
    def dataloader(self):
        """
        Get iterative dataloader

        Returns:
            :obj:`InterleaveDataset`:iterative tf.data.dataset
        """
        return self._dataloader



# Copyright 2020-present Tae Hwan Jung
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import unittest
from minio import Minio

from tests.test_storage import StorageTest

from matorage.nas import NAS

class DataSaverTest(StorageTest, unittest.TestCase):
    minio_config = Minio(
        endpoint="127.0.0.1:9001",
        access_key="minio",
        secret_key="miniosecretkey",
        secure=False,
    )
    nas_config = NAS(path="/tmp/unittest")

    def _create_bucket(self):
        self.minio_config.make_bucket('testminio')
        self.nas_config.make_bucket('testnas')

    def _get_obj_names(self, generator):
        res = sorted([g.object_name for g in generator])
        return res

    def _check_assertEqual(self, prefix=''):
        for r in [True, False]:
            self.assertEqual(
                self._get_obj_names(
                    self.minio_config.list_objects('testminio', prefix=prefix, recursive=r)
                ),
                self._get_obj_names(
                    self.nas_config.list_objects('testnas', prefix=prefix, recursive=r)
                )
            )

    def test_bucket_exists(self):
        self._create_bucket()
        self.assertEqual(
            self.minio_config.bucket_exists('testminio'),
            self.nas_config.bucket_exists('testnas')
        )

    def test_fput_object_with_list_objects(self):
        self._create_bucket()

        with open('test', 'w') as f:
            f.write('this is test')

        for _object_name in ['test', 'metadata/test']:
            self.minio_config.fput_object(
                bucket_name='testminio',
                object_name=_object_name,
                file_path='test'
            )
        for _object_name in ['test', 'metadata/test']:
            self.nas_config.fput_object(
                bucket_name='testnas',
                object_name=_object_name,
                file_path='test'
            )

        self._check_assertEqual(prefix='')
        self._check_assertEqual(prefix='metadata/')
        self._check_assertEqual(prefix='metadata')
        self._check_assertEqual(prefix='meta')

        self._check_assertEqual(prefix='metadata/t')
        self._check_assertEqual(prefix='metadata/test')

        self.assertEqual(
            self._get_obj_names(
                self.minio_config.list_objects('testminio', prefix='metadata/test/')
            ),
            self._get_obj_names(
                self.nas_config.list_objects('testnas', prefix='metadata/test/')
            )
        )

        self.assertNotEqual(
            self._get_obj_names(
                self.minio_config.list_objects('testminio', prefix='metadata/test/', recursive=True)
            ),
            self._get_obj_names(
                self.nas_config.list_objects('testnas', prefix='metadata/test/', recursive=True)
            )
        )

def suite():
    suties = unittest.TestSuite()
    suties.addTests(unittest.makeSuite(DataSaverTest))
    return suties


if __name__ == "__main__":
    unittest.main(defaultTest="suite")



# Copyright 2020-present Tae Hwan Jung
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import copy
import hashlib
from minio import Minio

from matorage.nas import NAS
from matorage.config import StorageConfig
from matorage.utils import check_nas, logger


class ModelConfig(StorageConfig):
    """
    Model configuration classes. This class overrides ``StorageConfig``.

    Args:
        endpoint (:obj:`string`, **require**):
            S3 object storage endpoint. or If use NAS setting, NAS folder path.
        access_key (:obj:`string`, optional, defaults to `None`):
            Access key for the object storage endpoint. (Optional if you need anonymous access).
        secret_key (:obj:`string`, optional, defaults to `None`):
            Secret key for the object storage endpoint. (Optional if you need anonymous access).
        secure (:obj:`boolean`, optional, defaults to `False`):
            Set this value to True to enable secure (HTTPS) access. (Optional defaults to False unlike the original MinIO).

        model_name (:obj:`string`, **require**):
            model name.
        additional (:obj:`dict`, optional, defaults to ``{}``):
            Parameters for additional description of models. The key and value of the dictionay can be specified very freely.
        compressor (:obj:`dict`, optional, defaults to :code:`{"complevel" : 0, "complib" : "zlib"}`):
            Model compressor option. It consists of a dict type that has complevel and complib as keys.
            For further reference, read `pytable's Filter <http://www.pytables.org/usersguide/libref/helper_classes.html#tables.Filters>`_.

            - complevel (:obj:`integer`, defaults to 0) : compressor level(0~9). The larger the number, the more compressed it is.
            - complib (:obj:`string`, defaults to 'zlib') : compressor library. choose in zlib, lzo, bzip2, blosc

    Examples::

        from matorage import ModelConfig

        model_config = ModelConfig(
            endpoint='127.0.0.1:9000',
            access_key='minio',
            secret_key='miniosecretkey',
            model_name='testmodel',
            additional={
                "version" : "1.0.1"
            }
        )

        model_config.to_json_file('testmodel.json')
        model_config2 = ModelConfig.from_json_file('testmodel.json')

    """

    def __init__(self, **kwargs):
        super(ModelConfig, self).__init__(**kwargs)
        self.type = "model"

        self.model_name = kwargs.pop("model_name", None)
        self.additional = kwargs.pop("additional", {})
        self.compressor = kwargs.pop("compressor", {"complevel": 0, "complib": "zlib"})

        self.bucket_name = self._hashmap_transfer()

        self.metadata = {
            "endpoint": self.endpoint,
            "model_name": self.model_name,
            "additional": self.additional,
            "compressor": self.compressor,
            "model": {},
        }

        self._check_all()

    def _check_all(self):
        """
        Check all class variable is fine.

        """
        self._check_bucket()

        if self.compressor["complevel"] < 0 or 9 < self.compressor["complevel"]:
            raise ValueError(
                "Compressor level is {} must be 0-9 interger".format(
                    self.compressor["level"]
                )
            )
        if self.compressor["complib"] not in ("zlib", "lzo", "bzip2", "blosc"):
            raise ValueError(
                "compressor mode {} is not valid. select in "
                "zlib, lzo, bzip2, blosc".format(self.compressor["lib"])
            )

    def _check_bucket(self):
        """
        Check bucket name is exist. If not exist, create new bucket
        If bucket and metadata sub folder exist, get metadata(attributes, compressor) from there.

        """
        _client = (
            Minio(
                self.endpoint,
                access_key=self.access_key,
                secret_key=self.secret_key,
                secure=self.secure,
                region=self.region
            )
            if not check_nas(self.endpoint)
            else NAS(self.endpoint)
        )
        if _client.bucket_exists(self.bucket_name):
            try:
                _metadata = _client.get_object(self.bucket_name, "metadata.json")
            except:
                _client.remove_bucket(self.bucket_name)
                raise FileNotFoundError(
                    "metadata.json is not in bucket name {}"
                    ", So this bucket will be removed".format(self.bucket_name)
                )

            metadata_dict = json.loads(_metadata.read().decode("utf-8"))
            if self.endpoint != metadata_dict["endpoint"]:
                raise ValueError(
                    "Already created endpoint({}) doesn't current endpoint str({})"
                    " It may occurs permission denied error".format(
                        metadata_dict["endpoint"], self.endpoint
                    )
                )

            self.compressor = metadata_dict["compressor"]
            self.metadata = metadata_dict
        else:
            logger.info(
                "{} {} is not exist!".format(self.model_name, str(self.additional))
            )

    def _hashmap_transfer(self):
        """
        Get unikey bucket name with `model_name` and `additional`

        Returns:
            :obj: `str`:
        """
        if not isinstance(self.model_name, str):
            raise ValueError(
                "model_name {} is empty or not str type".format(self.model_name)
            )
        if not isinstance(self.additional, dict):
            raise TypeError("additional is not dict type")

        key = (
            self.type
            + self.model_name
            + json.dumps(self.additional, indent=4, sort_keys=True)
        )
        return self.type + hashlib.md5(key.encode("utf-8")).hexdigest()

    def to_dict(self):
        """
        Serializes this instance to a Python dictionary.

        Returns:
            :obj:`Dict[str, any]`: Dictionary of all the attributes that make up this configuration instance,
        """
        output = copy.deepcopy(self.__class__.__base__(**self.__dict__).__dict__)
        output["type"] = self.type
        output["model_name"] = self.model_name
        output["additional"] = self.additional
        output["compressor"] = self.metadata["compressor"]
        return output



# Copyright 2020-present Tae Hwan Jung
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import json
import torch
import unittest
import numpy as np
from tqdm import tqdm
from torch.utils.data import DataLoader

from tests.test_data import DataTest

from matorage.data.config import DataConfig
from matorage.data.saver import DataSaver
from matorage.data.attribute import DataAttribute
from matorage.testing_utils import require_torch


@require_torch
class TorchDataTest(DataTest, unittest.TestCase):
    def test_torch_saver(self, data_config=None, save_to_json_file=False):
        if data_config is None:
            self.data_config = DataConfig(
                **self.storage_config,
                dataset_name="test_torch_saver",
                additional={"framework": "pytorch"},
                attributes=[
                    DataAttribute("image", "uint8", (2, 2), itemsize=32),
                    DataAttribute("target", "uint8", (1), itemsize=32),
                ]
            )
        else:
            self.data_config = data_config

        if save_to_json_file:
            self.data_config_file = "data_config_file.json"
            self.data_config.to_json_file(self.data_config_file)

        self.data_saver = DataSaver(config=self.data_config)

        self.data_saver(
            {
                "image": np.asarray([[[1, 2], [3, 4]], [[5, 6], [7, 8]]]),
                "target": np.asarray([0, 1]),
            }
        )
        self.data_saver.disconnect()

    def test_torch_loader(self):
        from matorage.torch import Dataset

        self.test_torch_saver()

        self.dataset = Dataset(config=self.data_config, cache_folder_path=self.cache_folder_path)
        loader = DataLoader(self.dataset, batch_size=64, num_workers=8, shuffle=True)

        for batch_idx, (image, target) in enumerate(tqdm(loader)):
            pass

    def test_torch_loader_with_compressor(self):
        from matorage.torch import Dataset

        data_config = DataConfig(
            **self.storage_config,
            dataset_name="test_torch_loader_with_compressor",
            additional={"framework": "pytorch"},
            compressor={"complevel": 4, "complib": "zlib"},
            attributes=[
                DataAttribute("image", "uint8", (2, 2), itemsize=32),
                DataAttribute("target", "uint8", (1), itemsize=32),
            ]
        )

        self.test_torch_saver(data_config=data_config)

        self.dataset = Dataset(config=data_config, cache_folder_path=self.cache_folder_path)
        loader = DataLoader(self.dataset, batch_size=64, num_workers=8, shuffle=True)

        for batch_idx, (image, target) in enumerate(tqdm(loader)):
            pass

    def test_torch_index(self):
        from matorage.torch import Dataset

        self.test_torch_saver()

        dataset = Dataset(config=self.data_config, index=True, cache_folder_path=self.cache_folder_path)

        assert torch.equal(
            dataset[0][0], torch.tensor([[1, 2], [3, 4]], dtype=torch.uint8)
        )
        assert torch.equal(dataset[0][1], torch.tensor([0], dtype=torch.uint8))

    def test_torch_index_with_compressor(self):
        from matorage.torch import Dataset

        data_config = DataConfig(
            **self.storage_config,
            dataset_name="test_torch_index_with_compressor",
            additional={"framework": "pytorch"},
            compressor={"complevel": 4, "complib": "zlib"},
            attributes=[
                DataAttribute("image", "uint8", (2, 2), itemsize=32),
                DataAttribute("target", "uint8", (1), itemsize=32),
            ]
        )

        self.test_torch_saver(data_config=data_config)

        dataset = Dataset(config=self.data_config, index=True, cache_folder_path=self.cache_folder_path)

        assert torch.equal(
            dataset[0][0], torch.tensor([[1, 2], [3, 4]], dtype=torch.uint8)
        )
        assert torch.equal(dataset[0][1], torch.tensor([0], dtype=torch.uint8))

    def test_saver_from_json_file(self):

        self.test_torch_saver(save_to_json_file=True)

        self.data_config = None
        self.data_saver = None

        self.data_config = DataConfig.from_json_file(self.data_config_file)

        self.data_saver = DataSaver(config=self.data_config)

        self.data_saver(
            {
                "image": np.asarray([[[1, 2], [3, 4]], [[5, 6], [7, 8]]]),
                "target": np.asarray([0, 1]),
            }
        )
        self.data_saver.disconnect()

    def test_loader_from_json_file(self):
        from matorage.torch import Dataset

        self.test_torch_saver(save_to_json_file=True)

        self.data_config = None

        self.data_config = DataConfig.from_json_file(self.data_config_file)

        self.dataset = Dataset(config=self.data_config, cache_folder_path=self.cache_folder_path)
        loader = DataLoader(self.dataset, batch_size=64, num_workers=8, shuffle=True)

        for batch_idx, (image, target) in enumerate(tqdm(loader)):
            pass

    def test_torch_not_clear(self):
        from matorage.torch import Dataset

        self.test_torch_loader()

        if os.path.exists(self.dataset.cache_path):
            with open(self.dataset.cache_path) as f:
                _pre_file_mapper = json.load(f)

        self.dataset = Dataset(config=self.data_config, clear=False, cache_folder_path=self.cache_folder_path)

        if os.path.exists(self.dataset.cache_path):
            with open(self.dataset.cache_path) as f:
                _next_file_mapper = json.load(f)

        self.assertEqual(_pre_file_mapper, _next_file_mapper)

    def test_datasaver_filetype(self):
        from matorage.torch import Dataset

        self.data_config = DataConfig(
            **self.storage_config,
            dataset_name="test_datasaver_filetype",
            attributes=[DataAttribute("x", "float64", (2), itemsize=32)],
        )
        self.data_saver = DataSaver(config=self.data_config)
        x = np.asarray([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]])
        self.assertEqual(x.shape, (3, 2))
        self.data_saver({"x": x})

        _file = open("test.txt", "w")
        _file.write('this is test')
        self.data_saver({"file": "test.txt"}, filetype=True)
        _file.close()

        self.data_saver.disconnect()

        self.dataset = Dataset(config=self.data_config, cache_folder_path=self.cache_folder_path)
        self.assertEqual(
            self.dataset.get_filetype_list, ["file"]
        )
        _local_filepath = self.dataset.get_filetype_from_key("file")
        with open(_local_filepath, 'r') as f:
            self.assertEqual(f.read(), 'this is test')

    def test_torch_saver_nas(self):
        self.data_config = DataConfig(
            **self.nas_config,
            dataset_name="test_torch_saver_nas",
            additional={"framework": "pytorch"},
            attributes=[
                DataAttribute("image", "uint8", (2, 2), itemsize=32),
                DataAttribute("target", "uint8", (1), itemsize=32),
            ]
        )

        self.data_saver = DataSaver(config=self.data_config)

        self.data_saver(
            {
                "image": np.asarray([[[1, 2], [3, 4]], [[5, 6], [7, 8]]]),
                "target": np.asarray([0, 1]),
            }
        )
        self.data_saver.disconnect()

    def test_torch_loader_nas(self):
        from matorage.torch import Dataset

        self.test_torch_saver_nas()

        self.dataset = Dataset(config=self.data_config, cache_folder_path=self.cache_folder_path)
        loader = DataLoader(self.dataset, batch_size=64, num_workers=8, shuffle=True)

        for batch_idx, (image, target) in enumerate(tqdm(loader)):
            pass

def suite():
    return unittest.TestSuite(unittest.makeSuite(TorchDataTest))


if __name__ == "__main__":
    unittest.main(defaultTest="suite")



# Copyright 2020-present Tae Hwan Jung
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import io
import json
import atexit
import bisect
import tempfile
from minio import Minio
from os.path import expanduser

from matorage.nas import NAS
from matorage.utils import logger, check_nas
from matorage.downloader import Downloader


class MTRData(object):
    r"""Parent Dataset class for Tensorflow and Pytorch Dataset

        This class is customized for the dataset of the Tensorflow, so it is operated by the following procedure.
        1. The `_object_file_mapper` manages the minio object as key and the downloaded local path as value.
            {'tmpv7sy5_1fff7845eccd874068.h5': '/tmp/tmpja6wo221tmpv7sy5_1fff7845eccd874068.h5'}
            When minio object is downloaded, it is recorded in _object_file_maper.
        2. We read `_object_file_mapper` and download only new objects that are not there.
        3. if Tensorflow v2(2.2.0>=), we use `tfio.IODataset.from_hdf5` and parallel `interleave` more fast

        Args:
            config (:obj:`matorage.DataConfig`, `require`):
            num_worker_threads :obj:`int`, `optional`, defaults to `4`):
                    number of backend storage worker to upload or download.
            clear (:obj:`boolean`, `optional`, defaults to `True`):
                Delete all files stored on the local storage after the program finishes.
            cache_folder_path (:obj:`str`, `optional`, defaults to `~/.matorage`):
                cached folder path to check which files are downloaded complete.
            index (:obj:`boolean`, `optional`, defaults to `False`):
                setting for index mode.
    """

    def __init__(
        self,
        config,
        num_worker_threads=4,
        clear=True,
        cache_folder_path="~/.matorage",
        index=False,
    ):
        self.config = config
        self.attribute = self._set_attribute()

        # Storage configuration
        self.num_worker_threads = num_worker_threads
        self.clear = clear
        self.index = index

        self._check_bucket()

        # merge all metadatas and load in memory.
        self.merged_indexer, self.merged_filetype = self._merge_metadata()
        self.end_indices = list(self.merged_indexer.keys())

        self._clients = {}

        if not self.index:
            # cache object which is downloaded.
            if not check_nas(self.config.endpoint):
                self._caching(cache_folder_path=cache_folder_path)
            else:
                self._object_file_mapper = {}

            # download all object in /tmp folder
            self._init_download()

            atexit.register(self._exit)

    def _caching(self, cache_folder_path):
        self.cache_folder_path = expanduser(cache_folder_path)
        if not os.path.exists(self.cache_folder_path):
            os.makedirs(self.cache_folder_path)

        self.cache_path = (
            f"{os.path.join(self.cache_folder_path, self.config.bucket_name)}.json"
        )
        if os.path.exists(self.cache_path):
            with open(self.cache_path) as f:
                self._object_file_mapper = json.load(f)
        else:
            self._object_file_mapper = {}

    def _check_bucket(self):
        _client = self._create_client()
        if not _client.bucket_exists(self.config.bucket_name):
            raise ValueError(
                "dataset {} with {} is not exist".format(
                    self.config.dataset_name, str(self.config.additional)
                )
            )

    def _create_client(self):
        return (
            Minio(
                endpoint=self.config.endpoint,
                access_key=self.config.access_key,
                secret_key=self.config.secret_key,
                secure=self.config.secure,
                region=self.config.region
            )
            if not check_nas(self.config.endpoint)
            else NAS(self.config.endpoint)
        )

    def _find_object(self, index):
        """
        find filename by index with binary search algorithm(indexes had been sorted).

        Returns:
            :obj:`str`: filename for index
        """
        _key_idx = bisect.bisect_right(self.end_indices, index)
        _key = self.end_indices[_key_idx]
        _last_key = self.end_indices[_key_idx - 1] if _key_idx else 0
        _relative_index = index - _last_key
        return self.merged_indexer[_key], _relative_index

    def _get_item_with_inmemory(self, idx):
        import h5py

        _pid = os.getpid()
        if _pid not in self._clients:
            self._clients[_pid] = self._create_client()

        _objectname, _relative_index = self._find_object(idx)
        _file_image = (
            self._clients[_pid]
            .get_object(self.config.bucket_name, object_name=_objectname)
            .read()
        )
        _file_image = h5py.File(io.BytesIO(_file_image), "r")

        return_tensor = {}
        for _attr_name in list(self.attribute.keys()):
            try:
                return_tensor[_attr_name] = self._reshape_convert_tensor(
                    numpy_array=_file_image[_attr_name][_relative_index],
                    attr_name=_attr_name,
                )
            except:
                raise IOError("Crash on concurrent read")

        return list(return_tensor.values())

    def _init_download(self):
        """
        Download all object from bucket with multi thread.
        cache to `_object_file_mapper` downloaded object paths.

        """
        _client = self._create_client()
        _downloader = Downloader(
            client=_client,
            bucket=self.config.bucket_name,
            num_worker_threads=self.num_worker_threads,
        )

        _remote_files = list(self.merged_indexer.values()) + list(self.merged_filetype)
        for _remote_file in _remote_files:
            if not check_nas(self.config.endpoint):
                _local_file = tempfile.mktemp(_remote_file)
                if _remote_file not in self._object_file_mapper:
                    self._object_file_mapper[_remote_file] = _local_file
                    _downloader.set_queue(
                        local_file=_local_file, remote_file=_remote_file
                    )
            else:
                if _remote_file not in self._object_file_mapper:
                    self._object_file_mapper[_remote_file] = os.path.join(
                        self.config.endpoint, self.config.bucket_name, _remote_file
                    )
        _downloader.join_queue()

        assert len(self._object_file_mapper) == (len(self.merged_indexer) + len(self.merged_filetype))

        if not check_nas(self.config.endpoint) and not os.path.exists(self.cache_path):
            with open(self.cache_path, "w") as f:
                json.dump(self._object_file_mapper, f)
            logger.info(
                "All {} {} datasets are downloaded done.".format(
                    self.config.dataset_name, str(self.config.additional)
                )
            )

    def _exit(self):
        """
        Close all opened files and remove.

        """

        if self.clear and not check_nas(self.config.endpoint):
            for _local_file in list(self._object_file_mapper.values()):
                if os.path.exists(_local_file):
                    os.remove(_local_file)
            if os.path.exists(self.cache_path):
                os.remove(self.cache_path)

    def _merge_metadata(self):
        """
        merge splited metadatas to a one file.
        +) merge dataset of filetype list

        Returns:
            :obj:`dict` : last end indexes with filename
            {
                3335: 'tmpajivq0tw0923909106de4222.h5',
                6670: 'tmp1g5zxyl0576b788d259844d1.h5',
                10005: 'tmpqnkklb9u27395376c94d4c14.h5'
            }

        """
        client = self._create_client()
        objects = client.list_objects(self.config.bucket_name, prefix="metadata/")

        total_index = []
        filetypes = []
        for obj in objects:
            metadata = client.get_object(
                self.config.bucket_name, object_name=obj.object_name
            )
            metadata = json.loads(metadata.read().decode("utf-8"))
            local_indexer = metadata["indexer"]
            total_index.extend(list(local_indexer.values()))
            filetypes.extend(metadata["filetype"])

        reindexer = {}
        for _index in total_index:
            key = (
                list(reindexer.keys())[-1] + _index["length"]
                if reindexer
                else _index["length"]
            )
            reindexer[key] = _index["name"]

        return reindexer, filetypes

    def _set_attribute(self):
        """
        Set `attribute` dictionary.

        Returns:
            :obj:`dict` : attribute
            {
                'image': {'shape': (28, 28), 'type': 'uint8'},
                'target': {'shape': (1,), 'type': 'uint8'}
            }
        """
        _attributes = {}
        _metadata_attributes = self.config.metadata.attributes
        for _attr in _metadata_attributes:
            _attributes[_attr.name] = {
                "shape": _attr.shape,
                "type": str(_attr.type.type),
            }
        return _attributes

    @property
    def get_filetype_list(self):
        """
        Get list of filetype dataset in bucket of ``DataConfig``

        Returns:
            :obj: `list`: list of key of filetype dataset in bucket of ``DataConfig``
        """
        return self.merged_filetype

    def get_filetype_from_key(self, filename):
        """
        Download filetype dataset from key in bucket of ``DataConfig``

        Args:
            filename (:obj:`string`):
                file name

        Returns:
            :obj: `string`: local path of downloaded file in bucket of ``DataConfig``
        """

        assert filename in self._object_file_mapper
        return self._object_file_mapper[filename]


# Copyright 2020-present Tae Hwan Jung
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from minio import Minio

from matorage.nas import NAS
from matorage.utils import check_nas, logger
from matorage.serialize import Serialize


class StorageConfig(Serialize):
    """
    Storage connector configuration classes.
    For MinIO, see `this page <https://docs.min.io/docs/python-client-api-reference.html>`_ for more details.

    Args:
        endpoint (:obj:`string`, **require**):
            S3 object storage endpoint. or If use NAS setting, NAS folder path.
        access_key (:obj:`string`, optional, defaults to `None`):
            Access key for the object storage endpoint. (Optional if you need anonymous access).
        secret_key (:obj:`string`, optional, defaults to `None`):
            Secret key for the object storage endpoint. (Optional if you need anonymous access).
        secure (:obj:`boolean`, optional, defaults to `False`):
            Set this value to True to enable secure (HTTPS) access. (Optional defaults to False unlike the original MinIO).

    Examples::

        from matorage import StorageConfig
        storage_config = StorageConfig(
            endpoint='127.0.0.1:9000',
            access_key='minio',
            secret_key='miniosecretkey'
        )


    """

    def __init__(self, **kwargs):

        # MinIO configuration
        self.endpoint = kwargs.pop("endpoint", None)
        self.access_key = kwargs.pop("access_key", None)
        self.secret_key = kwargs.pop("secret_key", None)
        self.secure = kwargs.pop("secure", False)
        self.region = kwargs.pop("region", None)

    def get_datasets(self):
        """
        Get all datasets from endpoint storage

        Returns:
            :obj:`list`:`dict` type of item(`dataset_name`, `additional`, `compressor`, `attributes`)
        """
        res = []

        _client = self._create_client()
        buckets = _client.list_buckets()
        for bucket in buckets:
            if bucket.name.startswith('dataset'):
                _metadata_names = _client.list_objects(bucket.name, prefix="metadata/")
                for _metadata_name in _metadata_names:
                    _metadata = _client.get_object(
                        bucket.name, _metadata_name.object_name
                    )
                    metadata_dict = json.loads(_metadata.read().decode("utf-8"))
                    res.append({
                        'dataset_name': metadata_dict['dataset_name'],
                        'additional': metadata_dict['additional'],
                        'compressor': metadata_dict['compressor'],
                        'attributes': metadata_dict['attributes'],
                    })
                    break
        return res

    def _get_type(self, type):
        res = []

        _client = self._create_client()
        buckets = _client.list_buckets()
        for bucket in buckets:
            if bucket.name.startswith(type):
                _metadata = _client.get_object(bucket.name, "metadata.json")
                metadata_dict = json.loads(_metadata.read().decode("utf-8"))
                res.append({
                    f'{type}_name': metadata_dict[f'{type}_name'],
                    'additional': metadata_dict['additional'],
                    'compressor': metadata_dict['compressor'],
                    type: metadata_dict[type],
                })

        return res

    def get_models(self):
        """
        Get all models from endpoint storage

        Returns:
            :obj:`list`:`dict` type of item(`model_name`, `additional`, `compressor`, `model`)
        """

        return self._get_type(type='model')

    def get_optimizers(self):
        """
        Get all optimizers from endpoint storage

        Returns:
            :obj:`list`:`dict` type of item(`model_name`, `additional`, `compressor`, `optimizer`)
        """

        return self._get_type(type='optimizer')

    def _create_client(self):
        return (
            Minio(
                endpoint=self.endpoint,
                access_key=self.access_key,
                secret_key=self.secret_key,
                secure=self.secure,
                region=self.region,
            )
            if not check_nas(self.endpoint)
            else NAS(self.endpoint)
        )



# Copyright 2020-present Tae Hwan Jung
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import unittest
import numpy as np

from tests.test_data import DataTest, DataS3Test

from matorage.data.config import DataConfig
from matorage.data.saver import DataSaver
from matorage.data.attribute import DataAttribute


class DataSaverTest(DataTest, unittest.TestCase):
    def test_dataconfig_one_attribute(self):
        self.data_config = DataConfig(
            **self.storage_config,
            dataset_name="test_dataconfig_one_attribute",
            attributes=DataAttribute("x", "uint8", (1)),
        )

    def test_dataconfig_one_attribute_with_tuple_attributes(self):
        self.data_config = DataConfig(
            **self.storage_config,
            dataset_name="test_dataconfig_one_attribute_with_tuple_attributes",
            attributes=("x", "uint8", (1)),
        )

    def test_reload_dataconfig(self):
        self.data_config = DataConfig(
            **self.storage_config,
            dataset_name="test_reload_dataconfig",
            attributes=DataAttribute("x", "uint8", (1)),
        )
        self.data_config_file = "data_config_file.json"
        self.data_config.to_json_file(self.data_config_file)

        self.data_config = None

        self.data_config = DataConfig.from_json_file(self.data_config_file)

    def test_dataconfig_two_attributes(self):
        self.data_config = DataConfig(
            **self.storage_config,
            dataset_name="test_dataconfig_two_attributes",
            attributes=[
                DataAttribute("x", "uint8", (1)),
                DataAttribute("y", "uint8", (1)),
            ],
        )

    def test_dataconfig_two_attribute_with_tuple_attributes(self):
        self.data_config = DataConfig(
            **self.storage_config,
            dataset_name="test_dataconfig_two_attribute_with_tuple_attributes",
            attributes=[("x", "uint8", (1)), ("y", "uint8", (1))],
        )

    def test_dataconfig_attributes_already_exist(self):
        with self.assertRaisesRegex(KeyError, "is already exist in"):
            self.data_config = DataConfig(
                **self.storage_config,
                dataset_name="test_dataconfig_attributes_already_exist",
                attributes=[
                    DataAttribute("x", "uint8", (1)),
                    DataAttribute("x", "uint8", (1)),
                ],
            )

    def test_dataconfig_string_attribute(self):
        self.data_config = DataConfig(
            **self.storage_config,
            dataset_name="test_dataconfig_string_attribute",
            attributes=[DataAttribute("x", "string", (1), itemsize=32)],
        )

    def test_dataconfig_bool_attribute(self):
        self.data_config = DataConfig(
            **self.storage_config,
            dataset_name="test_dataconfig_bool_attribute",
            attributes=[DataAttribute("x", "bool", (1))],
        )

    def test_dataconfig_int8_attribute(self):
        self.data_config = DataConfig(
            **self.storage_config,
            dataset_name="test_dataconfig_int8_attribute",
            attributes=[DataAttribute("x", "int8", (1))],
        )

    def test_dataconfig_int16_attribute(self):
        self.data_config = DataConfig(
            **self.storage_config,
            dataset_name="test_dataconfig_int16_attribute",
            attributes=[DataAttribute("x", "int16", (1))],
        )

    def test_dataconfig_int32_attribute(self):
        self.data_config = DataConfig(
            **self.storage_config,
            dataset_name="test_dataconfig_int32_attribute",
            attributes=[DataAttribute("x", "int32", (1))],
        )

    def test_dataconfig_uint8_attribute(self):
        self.data_config = DataConfig(
            **self.storage_config,
            dataset_name="test_dataconfig_uint8_attribute",
            attributes=[DataAttribute("x", "uint8", (1))],
        )

    def test_dataconfig_uint16_attribute(self):
        self.data_config = DataConfig(
            **self.storage_config,
            dataset_name="test_dataconfig_uint16_attribute",
            attributes=[DataAttribute("x", "uint16", (1))],
        )

    def test_dataconfig_uint32_attribute(self):
        self.data_config = DataConfig(
            **self.storage_config,
            dataset_name="test_dataconfig_uint32_attribute",
            attributes=[DataAttribute("x", "uint32", (1))],
        )

    def test_dataconfig_uint64_attribute(self):
        self.data_config = DataConfig(
            **self.storage_config,
            dataset_name="test_dataconfig_uint64_attribute",
            attributes=[DataAttribute("x", "uint64", (1))],
        )

    def test_dataconfig_float32_attribute(self):
        self.data_config = DataConfig(
            **self.storage_config,
            dataset_name="test_dataconfig_float32_attribute",
            attributes=[DataAttribute("x", "float32", (1))],
        )

    def test_dataconfig_float64_attribute(self):
        self.data_config = DataConfig(
            **self.storage_config,
            dataset_name="test_dataconfig_float64_attribute",
            attributes=[DataAttribute("x", "float64", (1))],
        )

    def test_datasaver_string_attribute(self):
        self.data_config = DataConfig(
            **self.storage_config,
            dataset_name="test_datasaver_string_attribute",
            attributes=[DataAttribute("x", "string", (2), itemsize=32)],
        )
        self.data_saver = DataSaver(config=self.data_config)
        x = np.asarray([["a", "b"], ["c", "d"], ["e", "f"]])
        self.assertEqual(x.shape, (3, 2))
        self.data_saver({"x": x})
        self.data_saver.disconnect()

    def test_datasaver_bool_attribute(self):
        self.data_config = DataConfig(
            **self.storage_config,
            dataset_name="test_datasaver_bool_attribute",
            attributes=[DataAttribute("x", "bool", (2), itemsize=32)],
        )
        self.data_saver = DataSaver(config=self.data_config)
        x = np.asarray([[True, False], [False, True], [True, True]])
        self.assertEqual(x.shape, (3, 2))
        self.data_saver({"x": x})
        self.data_saver.disconnect()

    def test_datasaver_int8_attribute(self):
        self.data_config = DataConfig(
            **self.storage_config,
            dataset_name="test_datasaver_int8_attribute",
            attributes=[DataAttribute("x", "int8", (2), itemsize=32)],
        )
        self.data_saver = DataSaver(config=self.data_config)
        x = np.asarray([[1, 2], [3, 4], [5, 6]])
        self.assertEqual(x.shape, (3, 2))
        self.data_saver({"x": x})
        self.data_saver.disconnect()

    def test_datasaver_int16_attribute(self):
        self.data_config = DataConfig(
            **self.storage_config,
            dataset_name="test_datasaver_int16_attribute",
            attributes=[DataAttribute("x", "int16", (2), itemsize=32)],
        )
        self.data_saver = DataSaver(config=self.data_config)
        x = np.asarray([[1, 2], [3, 4], [5, 6]])
        self.assertEqual(x.shape, (3, 2))
        self.data_saver({"x": x})
        self.data_saver.disconnect()

    def test_datasaver_int32_attribute(self):
        self.data_config = DataConfig(
            **self.storage_config,
            dataset_name="test_datasaver_int32_attribute",
            attributes=[DataAttribute("x", "int32", (2), itemsize=32)],
        )
        self.data_saver = DataSaver(config=self.data_config)
        x = np.asarray([[1, 2], [3, 4], [5, 6]])
        self.assertEqual(x.shape, (3, 2))
        self.data_saver({"x": x})
        self.data_saver.disconnect()

    def test_datasaver_int64_attribute(self):
        self.data_config = DataConfig(
            **self.storage_config,
            dataset_name="test_datasaver_int64_attribute",
            attributes=[DataAttribute("x", "int64", (2), itemsize=32)],
        )
        self.data_saver = DataSaver(config=self.data_config)
        x = np.asarray([[1, 2], [3, 4], [5, 6]])
        self.assertEqual(x.shape, (3, 2))
        self.data_saver({"x": x})
        self.data_saver.disconnect()

    def test_datasaver_uint8_attribute(self):
        self.data_config = DataConfig(
            **self.storage_config,
            dataset_name="test_datasaver_uint8_attribute",
            attributes=[DataAttribute("x", "uint8", (2), itemsize=32)],
        )
        self.data_saver = DataSaver(config=self.data_config)
        x = np.asarray([[1, 2], [3, 4], [5, 6]])
        self.assertEqual(x.shape, (3, 2))
        self.data_saver({"x": x})
        self.data_saver.disconnect()

    def test_datasaver_uint16_attribute(self):
        self.data_config = DataConfig(
            **self.storage_config,
            dataset_name="test_datasaver_uint16_attribute",
            attributes=[DataAttribute("x", "uint16", (2), itemsize=32)],
        )
        self.data_saver = DataSaver(config=self.data_config)
        x = np.asarray([[1, 2], [3, 4], [5, 6]])
        self.assertEqual(x.shape, (3, 2))
        self.data_saver({"x": x})
        self.data_saver.disconnect()

    def test_datasaver_uint32_attribute(self):
        self.data_config = DataConfig(
            **self.storage_config,
            dataset_name="test_datasaver_uint32_attribute",
            attributes=[DataAttribute("x", "uint32", (2), itemsize=32)],
        )
        self.data_saver = DataSaver(config=self.data_config)
        x = np.asarray([[1, 2], [3, 4], [5, 6]])
        self.assertEqual(x.shape, (3, 2))
        self.data_saver({"x": x})
        self.data_saver.disconnect()

    def test_datasaver_uint64_attribute(self):
        self.data_config = DataConfig(
            **self.storage_config,
            dataset_name="test_datasaver_uint64_attribute",
            attributes=[DataAttribute("x", "uint64", (2), itemsize=32)],
        )
        self.data_saver = DataSaver(config=self.data_config)
        x = np.asarray([[1, 2], [3, 4], [5, 6]])
        self.assertEqual(x.shape, (3, 2))
        self.data_saver({"x": x})
        self.data_saver.disconnect()

    def test_datasaver_float32_attribute(self):
        self.data_config = DataConfig(
            **self.storage_config,
            dataset_name="test_datasaver_float32_attribute",
            attributes=[DataAttribute("x", "float32", (2), itemsize=32)],
        )
        self.data_saver = DataSaver(config=self.data_config)
        x = np.asarray([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]])
        self.assertEqual(x.shape, (3, 2))
        self.data_saver({"x": x})
        self.data_saver.disconnect()

    def test_datasaver_float64_attribute(self):
        self.data_config = DataConfig(
            **self.storage_config,
            dataset_name="test_datasaver_float64_attribute",
            attributes=[DataAttribute("x", "float64", (2), itemsize=32)],
        )
        self.data_saver = DataSaver(config=self.data_config)
        x = np.asarray([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]])
        self.assertEqual(x.shape, (3, 2))
        self.data_saver({"x": x})
        self.data_saver.disconnect()

    def test_datasaver_inmemory(self):
        self.data_config = DataConfig(
            **self.storage_config,
            dataset_name="test_datasaver_inmemory",
            attributes=[DataAttribute("x", "float64", (2), itemsize=32)],
        )
        self.data_saver = DataSaver(config=self.data_config, inmemory=True)
        x = np.asarray([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]])
        self.assertEqual(x.shape, (3, 2))
        self.data_saver({"x": x})
        self.data_saver.disconnect()

    def test_datasaver_zlib(self):
        for level in range(10):
            self.data_config = DataConfig(
                **self.storage_config,
                dataset_name="test_datasaver_zlib",
                attributes=[DataAttribute("x", "float64", (2), itemsize=32)],
                compressor={"complevel": level, "complib": "zlib"},
            )
            self.data_saver = DataSaver(config=self.data_config)
            x = np.asarray([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]])
            self.assertEqual(x.shape, (3, 2))
            self.data_saver({"x": x})
            self.data_saver.disconnect()

    def test_datasaver_lzo(self):
        for level in range(10):
            self.data_config = DataConfig(
                **self.storage_config,
                dataset_name="test_datasaver_lzo",
                attributes=[DataAttribute("x", "float64", (2), itemsize=32)],
                compressor={"complevel": level, "complib": "lzo"},
            )
            self.data_saver = DataSaver(config=self.data_config)
            x = np.asarray([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]])
            self.assertEqual(x.shape, (3, 2))
            self.data_saver({"x": x})
            self.data_saver.disconnect()

    def test_datasaver_bzip2(self):
        for level in range(10):
            self.data_config = DataConfig(
                **self.storage_config,
                dataset_name="test_datasaver_bzip2",
                attributes=[DataAttribute("x", "float64", (2), itemsize=32)],
                compressor={"complevel": level, "complib": "bzip2"},
            )
            self.data_saver = DataSaver(config=self.data_config)
            x = np.asarray([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]])
            self.assertEqual(x.shape, (3, 2))
            self.data_saver({"x": x})
            self.data_saver.disconnect()

    def test_datasaver_blosc(self):
        for level in range(10):
            self.data_config = DataConfig(
                **self.storage_config,
                dataset_name="test_datasaver_blosc",
                attributes=[DataAttribute("x", "float64", (2), itemsize=32)],
                compressor={"complevel": level, "complib": "blosc"},
            )
            self.data_saver = DataSaver(config=self.data_config)
            x = np.asarray([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]])
            self.assertEqual(x.shape, (3, 2))
            self.data_saver({"x": x})
            self.data_saver.disconnect()

    def test_datasaver_nas(self):

        self.data_config = DataConfig(
            endpoint="/tmp/unittest",
            dataset_name="test_datasaver_nas",
            attributes=[DataAttribute("x", "float64", (2), itemsize=32)],
        )
        self.data_saver = DataSaver(config=self.data_config)
        x = np.asarray([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]])
        self.assertEqual(x.shape, (3, 2))
        self.data_saver({"x": x})
        self.data_saver.disconnect()

    def test_datasaver_refresh(self):

        self.data_config = DataConfig(
            **self.storage_config,
            dataset_name="test_datasaver_refresh",
            attributes=[DataAttribute("x", "float64", (2), itemsize=32)],
        )
        for refresh in [False, True]:
            self.data_saver = DataSaver(config=self.data_config, refresh=refresh)
            x = np.asarray([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]])
            self.assertEqual(x.shape, (3, 2))
            self.data_saver({"x": x})
            self.data_saver.disconnect()

    def test_datasaver_context_manager(self):
        self.data_config = DataConfig(
            **self.storage_config,
            dataset_name="test_datasaver_context_manager",
            attributes=[DataAttribute("x", "float64", (2), itemsize=32)],
        )

        x = np.asarray([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]])
        with DataSaver(config=self.data_config) as self.data_saver:
            self.assertEqual(x.shape, (3, 2))
            self.data_saver({"x": x})


@unittest.skipIf(
    'access_key' not in os.environ or 'secret_key' not in os.environ, 'S3 Skip'
)
class DataS3SaverTest(DataS3Test, unittest.TestCase):
    def test_datasaver_s3(self):
        self.storage_config = {
            'endpoint': 's3.us-east-1.amazonaws.com',
            'access_key': os.environ['access_key'],
            'secret_key': os.environ['secret_key'],
            'region': 'us-east-1',
            'secure': False,
        }

        self.data_config = DataConfig(
            **self.storage_config,
            dataset_name="test_datasaver_s3",
            attributes=[DataAttribute("x", "float64", (2), itemsize=32)],
        )
        self.data_saver = DataSaver(config=self.data_config)
        x = np.asarray([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]])
        self.assertEqual(x.shape, (3, 2))
        self.data_saver({"x": x})
        self.data_saver.disconnect()

    def test_datasaver_s3_filetype(self):
        from matorage.torch import Dataset

        self.storage_config = {
            'endpoint': 's3.us-east-1.amazonaws.com',
            'access_key': os.environ['access_key'],
            'secret_key': os.environ['secret_key'],
            'region': 'us-east-1',
            'secure': False,
        }

        self.data_config = DataConfig(
            **self.storage_config,
            dataset_name="test_datasaver_s3_filetype",
            attributes=[DataAttribute("x", "float64", (2), itemsize=32)],
        )
        self.data_saver = DataSaver(config=self.data_config)
        x = np.asarray([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]])
        self.assertEqual(x.shape, (3, 2))
        self.data_saver({"x": x})

        _file = open("test.txt", "w")
        _file.write('this is test')
        self.data_saver({"file": "test.txt"}, filetype=True)
        _file.close()

        self.data_saver.disconnect()

        self.dataset = Dataset(config=self.data_config, cache_folder_path=self.cache_folder_path)
        self.assertEqual(
            self.dataset.get_filetype_list, ["file"]
        )
        _local_filepath = self.dataset.get_filetype_from_key("file")
        with open(_local_filepath, 'r') as f:
            self.assertEqual(f.read(), 'this is test')


def suite():
    suties = unittest.TestSuite()
    suties.addTests(unittest.makeSuite(DataSaverTest))
    suties.addTests(unittest.makeSuite(DataS3SaverTest))
    return suties


if __name__ == "__main__":
    unittest.main(defaultTest="suite")



# Copyright 2020-present Tae Hwan Jung
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

_KB = 1024
"""The size of a Kilobyte in bytes"""

_MB = 1024 * _KB
"""The size of a Megabyte in bytes"""

import os
import io
import h5py
import numpy as np
import tensorflow as tf
from collections import OrderedDict, defaultdict
from tensorflow.python.keras import backend as K
from tensorflow.python.keras import __version__ as keras_version
from tensorflow.python.keras.saving.hdf5_format import preprocess_weights_for_loading

from matorage.model.manager import Manager


class ModelManager(Manager):

    """
    Model Manager Tensorflow classes. This class overrides ``Manager``.

    Note:
        Unlike Dataset, model weight is loaded entirely into cpu memory.
        Therefore, the `HDF5_CORE` driver using the memory option is default setting.

    Args:
        config (:obj:`matorage.ModelConfig`, **require**):
            A ModelConfig instance object
        num_worker_threads (:obj:`int`, optional, defaults to `4`):
            Number of backend storage worker to upload or download.
        multipart_upload_size (:obj:`integer`, optional, defaults to `5 * 1024 * 1024`):
            size of the incompletely uploaded object.
            You can sync files faster with `multipart upload in MinIO. <https://github.com/minio/minio-py/blob/master/minio/api.py#L1795>`_
            This is because MinIO clients use multi-threading, which improves IO speed more
            efficiently regardless of Python's Global Interpreter Lock(GIL).

    Examples::

        from matorage import ModelConfig
        from matorage.tensorflow import ModelManager
        from tensorflow.keras import layers, Sequential

        model_config = ModelConfig(
            endpoint='127.0.0.1:9000',
            access_key='minio',
            secret_key='miniosecretkey',
            model_name='testmodel',
            additional={
                "version" : "1.0.1"
            }
        )

        model_manager = ModelManager(config=model_config)

        model = Sequential([
            Sequential([
                layers.Dense(10),
                layers.ReLU()
            ]),
            Sequential([
                layers.Dense(10),
                layers.ReLU()
            ])
        ])
        model.build(input_shape=(None, 5))

        model_manager.save(model, step=100)

    """

    def __init__(self, config, num_worker_threads=4, multipart_upload_size=5 * _MB):
        super(ModelManager, self).__init__(
            config, num_worker_threads, multipart_upload_size
        )

    def _save_model(self, model_folder, model):
        for layer in model.weights:
            self._save_layer(model_folder, layer.name, layer.numpy())

    def _load_model(self, model_folder, layers, model):
        weight = OrderedDict()

        if isinstance(model, str):
            keys = [model]
        else:
            keys = [w.name for w in model.weights]

        for layer in layers:
            name = layer.object_name
            if name.find("/") > -1:
                name = name[name.find("/") + 1 :]

            if name in keys:
                layer_image = self._client.get_object(
                    bucket_name=self.config.bucket_name,
                    object_name=f"{model_folder}/{name}",
                ).read()

                layer_image = h5py.File(io.BytesIO(layer_image), "r")
                weight[name] = tf.convert_to_tensor(layer_image[self.type][:])

        if isinstance(model, str):
            return weight
        else:
            self._load_state_dict(model=model, weight_dict=weight)

    def _load_state_dict(self, model, weight_dict):
        original_keras_version = keras_version
        original_backend = K.backend()

        weight_value_tuples = []
        for k, layer in enumerate(model.layers):
            weight_names = [l.name for l in layer.weights]
            if len(weight_names) == 0:
                continue
            weight_values = [
                np.asarray(weight_dict[weight_name]) for weight_name in weight_names
            ]

            symbolic_weights = layer.trainable_weights + layer.non_trainable_weights
            weight_values = preprocess_weights_for_loading(
                layer, weight_values, original_keras_version, original_backend
            )

            if len(weight_values) != len(symbolic_weights):
                raise ValueError(
                    "Layer #"
                    + str(k)
                    + ' (named "'
                    + layer.name
                    + '" in the current model) was found to '
                    "correspond to layer " + layer.name + " in the save file. "
                    "However the new layer "
                    + layer.name
                    + " expects "
                    + str(len(symbolic_weights))
                    + " weights, but the saved weights have "
                    + str(len(weight_values))
                    + " elements."
                )
            weight_value_tuples += zip(symbolic_weights, weight_values)
        K.batch_set_value(weight_value_tuples)

    def save(self, model, **kwargs):
        """
        save weight of model

        Args:
            model (:obj:`tf.keras.Model`, **require**):
                Tensorflow model (``tf.keras.Model`` type)
            kwargs (:obj:`**kwargs`, optional):
                metadata about step or epoch for model.

        Examples::

            >>> model_manager.save(model, step=0)
            >>> model_manager.save(model, epoch=0)
            >>> model_manager.save(model, epoch=0, step=0)

        """
        super(ModelManager, self).save(model, **kwargs)

    def load(self, model, **kwargs):
        """
        load weight of model

        Args:
            model (:obj:`tf.keras.Model` or `string`, **require**):
                Tensorflow model(``tf.keras.Model`` type) or layer name(string type).
            kwargs (:obj:`**kwargs`, optional):
                metadata about step or epoch for model.

        Returns:
            :obj:`None or OrderedDict`: If ``model`` is tensorflow model, weight is loaded into the model and return None.
            however, If it is a string type with the name of the layer, it returns the weight of the OrderedDict type.

        Examples::

            # Load entire model
            >>> model_manager.load(model, step=0)

            # Load sub-layer model
            >>> submodel = Sequential([
            >>>   Sequential([
            >>>     layers.Dense(10),
            >>>     layers.ReLU()
            >>>   ])
            >>> ])
            >>> submodel.build(input_shape=(None, 5))
            >>> model_manager.load(submodel, step=0)

            >>> model_manager.load('sequential/dense/kernel:0', step=0)
            OrderedDict([('sequential/dense/kernel:0',
            <tf.Tensor: shape=(5, 784), dtype=float32, numpy=
            array([[ 0.08160326,  0.00161414, -0.00507049, ...,  0.02965256,
                    -0.07447692,  0.02029154],
                   [-0.06808375,  0.0112161 , -0.0640984 , ..., -0.05060118,
                    -0.03650254,  0.01808494],
                   [ 0.00063588,  0.00848304, -0.01014224, ...,  0.0616277 ,
                    -0.05507688,  0.02844934],
                   [-0.00206905,  0.04553737,  0.03098481, ..., -0.05891491,
                     0.0705805 , -0.03912991],
                   [ 0.04252511, -0.04907732, -0.07053198, ...,  0.00260394,
                     0.07418892, -0.0714546 ]], dtype=float32)>)])

        """
        return super(ModelManager, self).load(model, **kwargs)



# Copyright 2020-present Tae Hwan Jung
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

_KB = 1024
"""The size of a Kilobyte in bytes"""

_MB = 1024 * _KB
"""The size of a Megabyte in bytes"""

import io
import h5py
import torch
import numpy as np
from collections import defaultdict, OrderedDict
from torch.optim.optimizer import Optimizer

from matorage.optimizer.manager import Manager


class OptimizerManager(Manager):

    """
    Optimizer Manager Pytorch classes. This class overrides ``Manager``.

    Note:
        Unlike Dataset, optimizer weight is loaded entirely into cpu memory.
        Therefore, the `HDF5_CORE` driver using the memory option is default setting.

    Args:
        config (:obj:`matorage.OptimizerConfig`, **require**):
            A OptimizerConfig instance object
        num_worker_threads (:obj:`int`, optional, defaults to `4`):
            Number of backend storage worker to upload or download.
        multipart_upload_size (:obj:`integer`, optional, defaults to `5 * 1024 * 1024`):
            size of the incompletely uploaded object.
            You can sync files faster with `multipart upload in MinIO. <https://github.com/minio/minio-py/blob/master/minio/api.py#L1795>`_
            This is because MinIO clients use multi-threading, which improves IO speed more
            efficiently regardless of Python's Global Interpreter Lock(GIL).

    Examples::

        from matorage.torch import OptimizerManager
        optimizer_config = OptimizerConfig(
            endpoint='127.0.0.1:9000',
            access_key='minio',
            secret_key='miniosecretkey',
            optimizer_name='testoptimizer',
            additional={
                "version" : "1.0.1"
            }
        )

        optimizer_manager = OptimizerManager(config=optimizer_config)
        optimizer = optim.Adam(Model().parameters(), lr=0.01)
        optimizer_manager.save(optimizer, step=100)

    """

    def __init__(self, config, num_worker_threads=4, multipart_upload_size=5 * _MB):
        super(OptimizerManager, self).__init__(
            config, num_worker_threads, multipart_upload_size
        )

    def _get_step(self, optimizer):
        state = optimizer.state_dict()["state"]
        if state:
            step = list(state.values())[0]["step"]
            return step
        else:
            return None

    def _set_metadata(self, metadata, optimizer, step):
        assert isinstance(optimizer, Optimizer)
        optimizer = optimizer.state_dict()
        metadata["optimizer"].update(
            {
                str(step): {
                    "framework": "pytorch",
                    "param_groups": optimizer["param_groups"],
                }
            }
        )

    def _set_scheduler(self, metadata, scheduler, step):
        assert isinstance(scheduler, dict)
        metadata["scheduler"].update({
            str(step): scheduler
        })

    def _save_optimizer(self, step, optimizer):
        assert isinstance(optimizer, Optimizer)
        state = optimizer.state_dict()["state"]
        for param_name, param_dict in state.items():
            for param_dict_key, param_dict_value in param_dict.items():
                if torch.is_tensor(param_dict_value):
                    param_dict_value = param_dict_value.cpu().numpy()
                elif isinstance(param_dict_value, int):
                    param_dict_value = np.asarray([param_dict_value])
                self._save_param(
                    step, group=param_name, name=param_dict_key, weight=param_dict_value
                )

    def _load_optimizer(self, step, layers, optimizer):
        assert isinstance(optimizer, Optimizer)

        weight = OrderedDict()
        step = str(step)

        if step not in self.config.metadata["optimizer"]:
            raise KeyError(
                "Available only in {}".format(
                    list(self.config.metadata["optimizer"].keys())
                )
            )
        weight["param_groups"] = self.config.metadata["optimizer"][step]["param_groups"]
        weight["state"] = defaultdict(lambda: defaultdict())

        for layer in layers:
            name = layer.object_name
            layer_image = self._client.get_object(
                bucket_name=self.config.bucket_name, object_name=name
            ).read()

            layer_image = h5py.File(io.BytesIO(layer_image), "r")
            step, param_name, param_dict_key = name.split("/")
            if layer_image[self.type][:].shape == (1,):
                value = layer_image[self.type][0]
            else:
                value = torch.from_numpy(layer_image[self.type][:])
            weight["state"][param_name][param_dict_key] = value

        optimizer.load_state_dict(weight)

    def save(self, optimizer, scheduler=None):
        """
        save weight of optimizer

        Args:
            optimizer (:obj:`torch.optim`, **require**):
                Pytorch optimizer.

        Examples::

            >>> model = Model()
            >>> optimizer = optim.Adam(model.parameters(), lr=0.01)
            # model training...
            >>> optimizer_manager.save(optimizer)

        """
        if scheduler:
            scheduler = scheduler.state_dict()
        super(OptimizerManager, self).save(optimizer, scheduler=scheduler)

    def load(self, optimizer, step):
        """
        load weight of optimizer

        Args:
            optimizer (:obj:`torch.optim`, **require**):
                Pytorch optimizer.
            step (:obj:`integer`, **require**):
                optimizer step.

        Examples::

            >>> optimizer_manager = OptimizerManager(config=optimizer_config)
            >>> optimizer = optim.Adam(model.parameters(), lr=0.01)
            >>> optimizer_manager.load(optimizer, step=938)

        """
        super(OptimizerManager, self).load(optimizer, step)

    def load_with_scheduler(self, optimizer, scheduler, step):
        """
        load weight of optimizer and scheduler

        Args:
            optimizer (:obj:`torch.optim`, **require**):
                Pytorch optimizer.
            scheduler (:obj:`torch.optim.lr_scheduler`, **require**):
                Pytorch scheduler.
            step (:obj:`integer`, **require**):
                optimizer step.

        Examples::

            >>> optimizer_manager = OptimizerManager(config=optimizer_config)
            >>> optimizer = optim.Adam(model.parameters(), lr=0.01)
            >>> scheduler = StepLR(optimizer, step_size=30, gamma=0.1)
            >>> optimizer_manager.load_with_scheduler(optimizer, scheduler, step=938)

        """
        super(OptimizerManager, self).load(optimizer, step)

        step = str(step)
        if step in self.config.metadata["scheduler"]:
            scheduler.load_state_dict(self.config.metadata["scheduler"][step])
        else:
            raise KeyError(
                "Available only in {}".format(
                    list(self.config.metadata["scheduler"].keys())
                )
            )


# Copyright 2020-present Tae Hwan Jung
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

_KB = 1024
"""The size of a Kilobyte in bytes"""

_MB = 1024 * _KB
"""The size of a Megabyte in bytes"""

import copy
import json
import tables
import hashlib
import tempfile
from minio import Minio
from functools import reduce

from matorage.nas import NAS
from matorage.utils import check_nas, logger
from matorage.config import StorageConfig
from matorage.data.metadata import DataMetadata
from matorage.data.attribute import DataAttribute


class DataConfig(StorageConfig):
    """
    Dataset configuration classes. This class overrides ``StorageConfig``.

    Args:
        endpoint (:obj:`string`, **require**):
            S3 object storage endpoint. or If use NAS setting, NAS folder path.
        access_key (:obj:`string`, optional, defaults to `None`):
            Access key for the object storage endpoint. (Optional if you need anonymous access).
        secret_key (:obj:`string`, optional, defaults to `None`):
            Secret key for the object storage endpoint. (Optional if you need anonymous access).
        secure (:obj:`boolean`, optional, defaults to `False`):
            Set this value to True to enable secure (HTTPS) access. (Optional defaults to False unlike the original MinIO).
        max_object_size (:obj:`integer`, optional, defaults to `10MB`):
            One object file is divided into `max_object_size` and stored.

        dataset_name (:obj:`string`, **require**):
            dataset name.
        attributes (:obj:`list`, **require**):
            DataAttribute type of list for data attributes
        additional (:obj:`dict`, optional, defaults to ``{}``):
            Parameters for additional description of datasets. The key and value of the dictionay can be specified very freely.
        compressor (:obj:`dict`, optional, defaults to :code:`{"complevel" : 0, "complib" : "zlib"}`):
            Data compressor option. It consists of a dict type that has complevel and complib as keys.
            For further reference, read `pytable's Filter <http://www.pytables.org/usersguide/libref/helper_classes.html#tables.Filters>`_.

            - complevel (:obj:`integer`, defaults to 0) : compressor level(0~9). The larger the number, the more compressed it is.
            - complib (:obj:`string`, defaults to 'zlib') : compressor library. choose in zlib, lzo, bzip2, blosc
        max_object_size (:obj:`integer`, optional, defaults to `10MB`):
            One object file is divided into `max_object_size` and stored.

    Examples::

        from matorage import DataConfig, DataAttribute
        data_config = DataConfig(
            endpoint='127.0.0.1:9000',
            access_key='minio',
            secret_key='miniosecretkey',
            dataset_name='mnist',
            additional={
                "framework" : "pytorch",
                "mode" : "training"
            },
            compressor={
                "complevel" : 0,
                "complib" : "zlib"
            },
            attributes=[
                ('image', 'float32', (28, 28)),
                ('target', 'int64', (1, ))
            ]
        )

        data_config.to_json_file('data_config.json')
        data_config2 = DataConfig.from_json_file('data_config.json')

    If you have NAS(network access storage) settings, You can save/load faster by using the endpoint as a NAS folder path.

    Examples::

        from matorage import DataConfig

        # NAS example
        data_config = DataConfig(
            endpoint='~/shared',
            dataset_name='mnist',
            additional={
                "framework" : "pytorch",
                "mode" : "training"
            },
            compressor={
                "complevel" : 0,
                "complib" : "zlib"
            },
            attributes=[
                ('image', 'float32', (28, 28)),
                ('target', 'int64', (1, ))
            ]
        )


    """

    def __init__(self, **kwargs):
        super(DataConfig, self).__init__(**kwargs)
        self.type = "dataset"

        self.dataset_name = kwargs.pop("dataset_name", None)
        self.additional = kwargs.pop("additional", {})
        self.attributes = kwargs.pop("attributes", None)
        self.compressor = kwargs.pop("compressor", {"complevel": 0, "complib": "zlib"})
        self.max_object_size = kwargs.pop("max_object_size", 10 * _MB)

        self.bucket_name = self._hashmap_transfer()

        self._check_all()

        self.metadata = DataMetadata(**self.__dict__)

    def _check_all(self):
        """
        Check all class variable is fine.

        """
        self._check_bucket()

        if self.attributes is None:
            raise ValueError("attributes is empty")
        if isinstance(self.attributes, tuple):
            self.attributes = DataAttribute(
                name=self.attributes[0],
                type=self.attributes[1],
                shape=self.attributes[2],
            )
        if isinstance(self.attributes, DataAttribute):
            self.attributes = [self.attributes]

        for i, attr in enumerate(self.attributes):
            if isinstance(attr, tuple):
                self.attributes[i] = DataAttribute(attr[0], attr[1], attr[2])

        attribute_names = set()
        for attribute in self.attributes:
            assert isinstance(attribute.type, tables.atom.Atom)
            if attribute.name in attribute_names:
                raise KeyError(
                    "{} is already exist in {}".format(attribute.name, attribute_names)
                )
            else:
                attribute_names.add(attribute.name)

        # To convert `self.attributes`'s shape to be flatten
        self.flatten_attributes = copy.deepcopy(self.attributes)
        self._convert_type_flatten()

        if self.compressor["complevel"] < 0 or 9 < self.compressor["complevel"]:
            raise ValueError(
                "Compressor level is {} must be 0-9 interger".format(
                    self.compressor["level"]
                )
            )
        if self.compressor["complib"] not in ("zlib", "lzo", "bzip2", "blosc"):
            raise ValueError(
                "compressor mode {} is not valid. select in "
                "zlib, lzo, bzip2, blosc".format(self.compressor["lib"])
            )

    def _check_bucket(self):
        """
        Check bucket name is exist. If not exist, create new bucket
        If bucket and metadata sub folder exist, get metadata(attributes, compressor) from there.

        """

        aws_region_list = [
            'us-east-1',
            'us-east-2',
            'us-west-1',
            'us-west-2',
            'eu-west-1',
            'eu-west-2',
            'ca-central-1',
            'eu-central-1',
            'sa-east-1',
            'cn-north-1',
            'ap-southeast-1',
            'ap-southeast-2',
            'ap-northeast-1',
            'ap-northeast-2',
        ]

        if "amazonaws.com" in self.endpoint:
            if (self.region not in aws_region_list):
                raise AssertionError(
                    'AWS endpoint should has region argument from {}'.format(
                        aws_region_list
                    )
                )
            if f"s3.{self.region}.amazonaws.com" not in self.endpoint:
                raise AssertionError('Endpoint has to be {}'.format(
                        f"s3.{self.region}.amazonaws.com"
                    )
                )

        _client = (
            Minio(
                self.endpoint,
                access_key=self.access_key,
                secret_key=self.secret_key,
                secure=self.secure,
                region=self.region,
            )
            if not check_nas(self.endpoint)
            else NAS(self.endpoint)
        )
        if _client.bucket_exists(self.bucket_name):
            objects = _client.list_objects(self.bucket_name, prefix="metadata/")
            _metadata = None
            for obj in objects:
                _metadata = _client.get_object(self.bucket_name, obj.object_name)
                break
            if not _metadata:
                return

            metadata_dict = json.loads(_metadata.read().decode("utf-8"))
            if self.endpoint != metadata_dict["endpoint"]:
                raise ValueError(
                    "Already created endpoint({}) doesn't current endpoint str({})"
                    " It may occurs permission denied error".format(
                        metadata_dict["endpoint"], self.endpoint
                    )
                )

            self.compressor = metadata_dict["compressor"]
            self.attributes = [
                DataAttribute(**item) for item in metadata_dict["attributes"]
            ]
        else:
            logger.info(
                "{} {} is not exist!".format(self.dataset_name, str(self.additional))
            )

    def _convert_type_flatten(self):
        for attribute in self.flatten_attributes:
            attribute.shape = (reduce(lambda x, y: x * y, attribute.shape),)

    def _hashmap_transfer(self):
        """
        Get unikey bucket name with `dataset_name` and `additional`

        Returns:
            :obj: `str`:
        """
        if not isinstance(self.dataset_name, str):
            raise ValueError(
                "dataset_name {} is empty or not str type".format(self.dataset_name)
            )
        if not isinstance(self.additional, dict):
            raise TypeError("additional is not dict type")

        key = self.dataset_name + json.dumps(self.additional, indent=4, sort_keys=True)
        return self.type + hashlib.md5(key.encode("utf-8")).hexdigest()

    def to_dict(self):
        """
        Serializes this instance to a Python dictionary.

        Returns:
            :obj:`Dict[str, any]`: Dictionary of all the attributes that make up this configuration instance,
        """
        output = copy.deepcopy(self.__class__.__base__(**self.__dict__).__dict__)
        output["dataset_name"] = self.metadata.dataset_name
        output["additional"] = self.metadata.additional
        output["attributes"] = [_attribute.to_dict() for _attribute in self.attributes]
        output["compressor"] = self.metadata.compressor
        return output

    @classmethod
    def from_json_file(cls, json_file):
        """
        Constructs a `Config` from the path to a json file of parameters.

        Args:
            json_file (:obj:`string`):
                Path to the JSON file containing the parameters.

        Returns:
            :class:`DataConfig`: An instance of a configuration object

        """
        config_dict = cls._dict_from_json_file(json_file)

        config_dict["attributes"] = [
            DataAttribute(**item) for item in config_dict["attributes"]
        ]

        return cls(**config_dict)

    def set_indexer(self, index):
        self.metadata.indexer.update(index)

    def set_files(self, files):
        self.metadata.filetype.append(files)

    @property
    def get_length(self):
        """
        Get length of dataset in ``DataConfig``

        Returns:
            :obj:`integer`: length of dataset

        """
        keys = list(self.metadata.indexer.keys())
        if len(keys) == 0:
            return 0
        else:
            return keys[-1]


# Copyright 2020-present Tae Hwan Jung
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import gc
import unittest
import tensorflow as tf
from tensorflow import keras
from unittest.case import SkipTest

from tests.test_model import ModelTest

from matorage.model.config import ModelConfig
from matorage.model.tensorflow.v2.manager import ModelManager
from matorage.testing_utils import require_tf


@require_tf
class TFModelTest(ModelTest, unittest.TestCase):
    flag = False

    def create_model(self):
        model = tf.keras.models.Sequential(
            [
                keras.layers.Dense(512, activation="relu", input_shape=(784,)),
                keras.layers.Dropout(0.2),
                keras.layers.Dense(10),
            ]
        )
        model.compile(
            optimizer="adam",
            loss=tf.losses.SparseCategoricalCrossentropy(from_logits=True),
            metrics=["accuracy"],
        )
        return model

    def test_tfmodel_saver(self, model_config=None, save_to_json_file=False):

        self.model = self.create_model()

        if model_config is None:
            self.model_config = ModelConfig(
                **self.storage_config,
                model_name="test_tfmodel_saver",
                additional={"framework": "tensorflow"}
            )
        else:
            self.model_config = model_config

        if save_to_json_file:
            self.model_config_file = "model_config_file.json"
            self.model_config.to_json_file(self.model_config_file)

        self.model_manager = ModelManager(config=self.model_config)

        self.model_manager.save(self.model, step=0)

    def test_tfmodel_saver_from_json_file(self):

        self.test_tfmodel_saver(save_to_json_file=True)

        self.model_config = None
        self.model_manager = None

        self.model_config = ModelConfig.from_json_file(self.model_config_file)

        self.model_manager = ModelManager(config=self.model_config)

        self.model_manager.save(self.model, step=0)

def suite():
    return unittest.TestSuite(unittest.makeSuite(TFModelTest))


if __name__ == "__main__":
    unittest.main(defaultTest="suite")



# Copyright 2020-present Tae Hwan Jung
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import json
import tables
import hashlib
import tempfile
from minio import Minio

from matorage.nas import NAS
from matorage.utils import check_nas, logger
from matorage.uploader import Uploader

_KB = 1024
"""The size of a Kilobyte in bytes"""

_MB = 1024 * _KB
"""The size of a Megabyte in bytes"""


class Manager(object):
    type = "optimizer"

    def __init__(self, config, num_worker_threads=4, multipart_upload_size=5 * _MB):
        self.config = config
        self.num_worker_threads = num_worker_threads
        self.multipart_upload_size = multipart_upload_size

        self._client = (
            Minio(
                endpoint=self.config.endpoint,
                access_key=self.config.access_key,
                secret_key=self.config.secret_key,
                secure=self.config.secure,
                region=self.config.region,
            )
            if not check_nas(self.config.endpoint)
            else NAS(self.config.endpoint)
        )

        self._uploader = Uploader(
            client=self._client,
            bucket=self.config.bucket_name,
            num_worker_threads=self.num_worker_threads,
            multipart_upload_size=self.multipart_upload_size,
            inmemory=True,
        )

    def _uploader_closing(self):
        self._uploader.join_queue()

        _metadata_file = tempfile.mktemp("metadata.json")
        with open(_metadata_file, "w", encoding="utf-8") as writer:
            writer.write(json.dumps(self.config.metadata, indent=4) + "\n")

        self._client.fput_object(
            bucket_name=self.config.bucket_name,
            object_name="metadata.json",
            file_path=_metadata_file,
        )
        os.remove(_metadata_file)

    def _save_with_clear(self, step, optimizer, overwrite=False):
        if overwrite:
            objects = self._client.list_objects(
                bucket_name=self.config.bucket_name, prefix=f"{step}/"
            )
            for obj in objects:
                self._client.remove_object(
                    bucket_name=self.config.bucket_name, object_name=obj.object_name
                )

        # saving optimizer
        self._save_optimizer(step, optimizer)
        self._uploader_closing()

    def _save_param(self, step, group, name, weight):
        group = str(group)
        _local_file = tempfile.mktemp(f"{name}.h5")

        _file = tables.open_file(
            _local_file, "w", driver="H5FD_CORE", driver_core_backing_store=False
        )
        _file.create_carray(
            "/", self.type, obj=weight, filters=tables.Filters(**self.config.compressor)
        )

        if group:
            self._uploader.set_queue(
                local_file=_file.get_file_image(), remote_file=f"{step}/{group}/{name}"
            )
        else:
            self._uploader.set_queue(
                local_file=_file.get_file_image(), remote_file=f"{step}/{name}"
            )
        _file.close()

    def save(self, optimizer, scheduler=None):
        if not self._client.bucket_exists(self.config.bucket_name):
            self._client.make_bucket(
                self.config.bucket_name, location=self.config.region
            )

        step = self._get_step(optimizer)
        if not step:
            logger.error(
                "{} {} step({})is not exist".format(
                    self.config.optimizer_name, self.config.additional, str(step)
                )
            )
            return

        if step in self.config.metadata["optimizer"]:
            logger.info(
                "{} {} is already exist, so optimizer will be overwrited.".format(
                    self.config.optimizer_name, str(self.config.additional)
                )
            )
            self._save_with_clear(step, optimizer, overwrite=True)
        else:
            self._set_metadata(
                metadata=self.config.metadata, optimizer=optimizer, step=step
            )
            self._save_with_clear(step, optimizer)

        if scheduler:
            self._set_scheduler(
                metadata=self.config.metadata, scheduler=scheduler, step=step
            )

        logger.info("optimizer with {} is saved".format(str(step)))

    def load(self, optimizer, step):
        layers = self._client.list_objects(
            bucket_name=self.config.bucket_name, prefix=f"{step}/", recursive=True
        )

        logger.info("optimizer with {} is loaded".format(str(step)))
        self._load_optimizer(step, layers, optimizer)

    @property
    def get_metadata(self):
        """
        Get all optimizers according to metadata by step.

        Returns:
            :obj:`dict`: optimizer of metadata

        Examples::

            >>> optimizer_manager = OptimizerManager(config=optimizer_config)
            >>> optimizer_manager.save(optimizer)
            >>> optimizer_manager.get_metadata
            {'938':
                {
                    'framework': 'pytorch',
                    'param_groups': [
                        {
                            'lr': 0.01, 'betas': [0.9, 0.999], 'eps': 1e-08,
                            'weight_decay': 0, 'amsgrad': False,
                            'params': [
                                140516594711520, 140516594711760,
                                140517867028384, 140516594711680,
                                140516594693376, 140516594612336
                            ]
                        }
                    ]
                }
            }

        """
        return self.config.metadata["optimizer"]



# Copyright 2020-present Tae Hwan Jung
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
from glob import iglob


class Obj(object):
    def __init__(self, object_name):
        self.object_name = object_name


class NAS(object):
    def __init__(self, path):
        self.path = path

    def bucket_exists(self, bucket_name):
        return os.path.exists(os.path.join(self.path, bucket_name))

    def fget_object(self, bucket_name, object_name, file_path):
        raise NotImplementedError

    def fput_object(self, bucket_name, object_name, file_path, part_size=None):
        _filename = os.path.join(self.path, bucket_name, object_name)
        os.makedirs(os.path.dirname(_filename), exist_ok=True)
        shutil.copyfile(src=file_path, dst=_filename)

    def get_object(self, bucket_name, object_name):
        _filename = os.path.join(self.path, bucket_name, object_name)
        return open(_filename, "rb")

    def put_object(self, bucket_name, object_name, data, length, part_size=None):
        _filename = os.path.join(self.path, bucket_name, object_name)
        os.makedirs(os.path.dirname(_filename), exist_ok=True)
        data.seek(0)
        with open(_filename, "wb") as f:
            shutil.copyfileobj(data, f, length=length)

    def list_objects(self, bucket_name, prefix="", recursive=False):
        _base = os.path.join(self.path, bucket_name)
        _query = os.path.join(self.path, bucket_name, prefix)
        objects = []

        if '/' in _query:
            _query = _query.rsplit('/', 1)[0] + '/'
        else:
            _query = _base

        for f in iglob(f"{_query}/**", recursive=recursive):
            if os.path.isdir(f):
                f = f.replace(f"{_base}/", '')
                if f and not recursive:
                    objects.append(f"{f}/")
            else:
                f = f.replace(f"{_base}/", '')
                if f:
                    objects.append(f)

        return [Obj(o) for o in objects if o.startswith(prefix)]

    def make_bucket(self, bucket_name, location='us-east-1'):
        os.makedirs(os.path.join(self.path, bucket_name))

    def remove_bucket(self, bucket_name):
        shutil.rmtree(os.path.join(self.path, bucket_name))

    def remove_object(self, bucket_name, object_name):
        _path = os.path.join(self.path, bucket_name, object_name)
        if os.path.isdir(_path):
            shutil.rmtree(_path)
        else:
            os.remove(_path)


# Copyright 2020-present Tae Hwan Jung
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
import unittest
import torch.nn as nn
from tqdm import tqdm
import torch.optim as optim
import torch.nn.functional as F
from torch.utils.data import DataLoader
from torchvision import datasets, transforms

from tests.test_optimizer import OptimizerTest

from matorage.optimizer.config import OptimizerConfig
from matorage.optimizer.torch.manager import OptimizerManager
from matorage.testing_utils import require_torch


@require_torch
class Model(nn.Module):
    def __init__(self):
        super(Model, self).__init__()
        self.fc1 = nn.Linear(28 * 28, 512)
        self.fc2 = nn.Linear(512, 256)
        self.fc3 = nn.Linear(256, 10)

    def forward(self, x):
        x = x.view(-1, 28 * 28)
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x


@require_torch
class TorchOptimizerTest(OptimizerTest, unittest.TestCase):
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    transform = transforms.Compose(
        [transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,))]
    )

    train_dataset = datasets.MNIST(
        "/tmp/data", train=True, download=True, transform=transform
    )

    def test_optimizer_saver(self, optimizer_config=None, save_to_json_file=False):

        model = Model().to(self.device)
        optimizer = optim.Adam(model.parameters(), lr=0.01)
        criterion = torch.nn.CrossEntropyLoss()

        train_loader = DataLoader(self.train_dataset, batch_size=64, num_workers=4)

        for batch_idx, (image, target) in enumerate(tqdm(train_loader)):
            image, target = image.to(self.device), target.to(self.device)
            optimizer.zero_grad()
            output = model(image)
            loss = criterion(output, target)
            loss.backward()
            optimizer.step()

        if optimizer_config is None:
            self.optimizer_config = OptimizerConfig(
                **self.storage_config,
                optimizer_name="testoptimizer",
                additional={"version": "1.0.1"}
            )
        else:
            self.optimizer_config = optimizer_config

        if save_to_json_file:
            self.optimizer_config_file = "optimizer_config_file.json"
            self.optimizer_config.to_json_file(self.optimizer_config_file)

        self.optimizer_manager = OptimizerManager(config=self.optimizer_config)
        self.optimizer_manager.save(optimizer)

    def test_optimizer_saver_from_json_file(self):

        self.test_optimizer_saver(save_to_json_file=True)

        self.optimizer_config = None
        self.optimizer_manager = None

        self.optimizer_config = OptimizerConfig.from_json_file(
            self.optimizer_config_file
        )

        self.optimizer_manager = OptimizerManager(config=self.optimizer_config)

        model = Model().to(self.device)
        optimizer = optim.Adam(model.parameters(), lr=0.01)
        criterion = torch.nn.CrossEntropyLoss()

        train_loader = DataLoader(self.train_dataset, batch_size=64, num_workers=4)

        for batch_idx, (image, target) in enumerate(tqdm(train_loader)):
            image, target = image.to(self.device), target.to(self.device)
            optimizer.zero_grad()
            output = model(image)
            loss = criterion(output, target)
            loss.backward()
            optimizer.step()

        self.optimizer_manager = OptimizerManager(config=self.optimizer_config)
        self.optimizer_manager.save(optimizer)

    def test_optimizer_loader(self):

        self.test_optimizer_saver()

        model = Model().to(self.device)
        optimizer = optim.Adam(model.parameters(), lr=0.01)

        self.optimizer_manager.load(optimizer, step=938)

    def test_optimizer_scheduler_saver(self):
        from torch.optim.lr_scheduler import StepLR

        model = Model().to(self.device)
        optimizer = optim.Adam(model.parameters(), lr=1.0)
        criterion = torch.nn.CrossEntropyLoss()
        scheduler = StepLR(optimizer, step_size=134, gamma=0.99)

        train_loader = DataLoader(self.train_dataset, batch_size=64, num_workers=4)

        for batch_idx, (image, target) in enumerate(tqdm(train_loader)):
            image, target = image.to(self.device), target.to(self.device)
            optimizer.zero_grad()
            output = model(image)
            loss = criterion(output, target)
            loss.backward()
            optimizer.step()
            scheduler.step()

        self.optimizer_config = OptimizerConfig(
            **self.storage_config,
            optimizer_name="testoptimizerwithscheduler",
            additional={"version": "1.0.1"}
        )
        self.optimizer_manager = OptimizerManager(config=self.optimizer_config)
        self.optimizer_manager.save(optimizer, scheduler)

        return scheduler

    def test_optimizer_scheduler_loader(self):
        from torch.optim.lr_scheduler import StepLR

        _scheduler = self.test_optimizer_scheduler_saver()

        model = Model()
        optimizer = optim.Adam(model.parameters(), lr=0.01)
        scheduler = StepLR(optimizer, step_size=30, gamma=0.1)

        self.optimizer_manager.load_with_scheduler(optimizer, scheduler, step=938)
        self.assertEqual(_scheduler.state_dict(), scheduler.state_dict())

    def test_optimizer_saver_nas(self):

        model = Model().to(self.device)
        optimizer = optim.Adam(model.parameters(), lr=0.01)
        criterion = torch.nn.CrossEntropyLoss()

        train_loader = DataLoader(self.train_dataset, batch_size=64, num_workers=4)

        for batch_idx, (image, target) in enumerate(tqdm(train_loader)):
            image, target = image.to(self.device), target.to(self.device)
            optimizer.zero_grad()
            output = model(image)
            loss = criterion(output, target)
            loss.backward()
            optimizer.step()

        self.optimizer_config = OptimizerConfig(
            **self.nas_config,
            optimizer_name="testoptimizernas",
            additional={"version": "1.0.1"}
        )

        self.optimizer_manager = OptimizerManager(config=self.optimizer_config)
        self.optimizer_manager.save(optimizer)

    def test_optimizer_loader_nas(self):

        self.test_optimizer_saver_nas()

        model = Model().to(self.device)
        optimizer = optim.Adam(model.parameters(), lr=0.01)

        self.optimizer_manager.load(optimizer, step=938)

def suite():
    return unittest.TestSuite(unittest.makeSuite(TorchOptimizerTest))


if __name__ == "__main__":
    unittest.main(defaultTest="suite")



# Copyright 2020-present Tae Hwan Jung
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
import unittest
import numpy as np
from tqdm import tqdm
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.data import DataLoader
from torchvision import datasets, transforms
from unittest.case import SkipTest

from tests.test_model import ModelTest

from matorage.model.config import ModelConfig
from matorage.model.torch.manager import ModelManager
from matorage.testing_utils import require_torch


@require_torch
class Model(nn.Module):
    def __init__(self):
        super(Model, self).__init__()
        self.fc1 = nn.Linear(28 * 28, 512)
        self.fc2 = nn.Linear(512, 256)
        self.fc3 = nn.Linear(256, 10)

    def forward(self, x):
        x = x.view(-1, 28 * 28)
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x


@require_torch
class TorchModelTest(ModelTest, unittest.TestCase):

    transform = transforms.Compose(
        [transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,))]
    )
    flag = False

    def test_torchmodel_saver(self, model_config=None, save_to_json_file=False):
        if model_config is None:
            self.model_config = ModelConfig(
                **self.storage_config,
                model_name="test_torchmodel_saver",
                additional={"framework": "pytorch"}
            )
        else:
            self.model_config = model_config

        if save_to_json_file:
            self.model_config_file = "model_config_file.json"
            self.model_config.to_json_file(self.model_config_file)

        self.model_manager = ModelManager(config=self.model_config)

        model = Model()
        self.model_manager.save(model, step=0)

    def test_torchmodel_saver_from_json_file(self):

        self.test_torchmodel_saver(save_to_json_file=True)

        self.model_config = None
        self.model_manager = None

        self.model_config = ModelConfig.from_json_file(self.model_config_file)

        self.model_manager = ModelManager(config=self.model_config)

        model = Model()
        self.model_manager.save(model, step=0)

    def test_torchmodel_loader(self):

        self.test_torchmodel_saver()

        model = Model()
        self.model_manager.load(model, step=0)

    def test_torchmodel_loader_with_compressor(self):

        model_config = ModelConfig(
            **self.storage_config,
            model_name="test_torchmodel_loader_with_compressor",
            additional={"framework": "pytorch"},
            compressor={"complevel": 4, "complib": "zlib"}
        )

        self.test_torchmodel_saver(model_config=model_config)

        self.model_manager = ModelManager(config=self.model_config)

        model = Model()
        self.model_manager.load(model, step=0)

    def test_torchmodel_layer_loader(self):

        self.test_torchmodel_saver()

        self.model_manager = ModelManager(config=self.model_config)

        self.model_manager.load("f.weight", step=0)

    def test_mnist_eval(self, model=None, device=None):
        if not self.flag:
            raise SkipTest

        test_dataset = datasets.MNIST(
            "/tmp/data", train=False, transform=self.transform
        )
        test_loader = DataLoader(test_dataset, batch_size=64, num_workers=4)

        model.eval()
        test_loss = 0
        correct = 0
        with torch.no_grad():
            for image, target in test_loader:
                image, target = image.to(device), target.to(device)
                output = model(image)
                test_loss += F.nll_loss(output, target, reduction="sum").item()
                pred = output.argmax(dim=1, keepdim=True)
                correct += pred.eq(target.view_as(pred)).sum().item()

        test_loss /= len(test_loader.dataset)
        return correct

    def test_mnist_reloaded(self):
        self.flag = True

        import torch.optim as optim

        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        train_dataset = datasets.MNIST(
            "/tmp/data", train=True, download=True, transform=self.transform
        )

        model = Model().to(device)
        optimizer = optim.Adam(model.parameters(), lr=0.01)
        criterion = torch.nn.CrossEntropyLoss()

        train_loader = DataLoader(train_dataset, batch_size=64, num_workers=4)

        for batch_idx, (image, target) in enumerate(tqdm(train_loader)):
            image, target = image.to(device), target.to(device)
            optimizer.zero_grad()
            output = model(image)
            loss = criterion(output, target)
            loss.backward()
            optimizer.step()

        self.model_config = ModelConfig(
            **self.storage_config,
            model_name="testmodel",
            additional={"version": "1.0.1"}
        )
        self.model_manager = ModelManager(config=self.model_config)

        self.model_manager.save(model, epoch=1)

        pretrained_model = Model().to(device)
        correct = self.test_mnist_eval(model=pretrained_model, device=device)

        self.model_manager.load(pretrained_model, epoch=1)
        pretrained_correct = self.test_mnist_eval(model=pretrained_model, device=device)

        assert correct < pretrained_correct

    def test_mnist_reloaded_nas(self):
        self.flag = True

        import torch.optim as optim

        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        train_dataset = datasets.MNIST(
            "/tmp/data", train=True, download=True, transform=self.transform
        )

        model = Model().to(device)
        optimizer = optim.Adam(model.parameters(), lr=0.01)
        criterion = torch.nn.CrossEntropyLoss()

        train_loader = DataLoader(train_dataset, batch_size=64, num_workers=4)

        for batch_idx, (image, target) in enumerate(tqdm(train_loader)):
            image, target = image.to(device), target.to(device)
            optimizer.zero_grad()
            output = model(image)
            loss = criterion(output, target)
            loss.backward()
            optimizer.step()

        self.model_config = ModelConfig(
            **self.nas_config,
            model_name="test_torch_mnist_nas",
            additional={"version": "1.0.1"}
        )
        self.model_manager = ModelManager(config=self.model_config)

        self.model_manager.save(model, epoch=1)

        pretrained_model = Model().to(device)
        correct = self.test_mnist_eval(model=pretrained_model, device=device)

        self.model_manager.load(pretrained_model, epoch=1)
        pretrained_correct = self.test_mnist_eval(model=pretrained_model, device=device)

        assert correct < pretrained_correct


def suite():
    return unittest.TestSuite(unittest.makeSuite(TorchModelTest))


if __name__ == "__main__":
    unittest.main(defaultTest="suite")



# Copyright 2020-present Tae Hwan Jung
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from matorage.serialize import Serialize

from matorage.data.config import DataConfig
from matorage.data.attribute import DataAttribute
from matorage.data.saver import DataSaver

from matorage.config import StorageConfig

from matorage.model.config import ModelConfig
from matorage.optimizer.config import OptimizerConfig

__all__ = [
    "Serialize",
    "StorageConfig",
    "DataAttribute",
    "DataConfig",
    "DataSaver",
    "ModelConfig",
    "OptimizerConfig",
]



# Original Code
# https://github.com/huggingface/transformers/blob/master/src/transformers/file_utils.py
# See https://github.com/graykode/matorage/blob/0.1.0/NOTICE
# modified by TaeHwan Jung(@graykode)

# Copyright 2020-present Tae Hwan Jung
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging
from packaging import version
from urllib.parse import urlsplit

# from https://github.com/huggingface/transformers/blob/master/examples/distillation/utils.py
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logging.basicConfig(
    format="%(asctime)s - %(levelname)s - %(name)s - PID: %(process)d -  %(message)s",
    datefmt="%m/%d/%Y %H:%M:%S",
    level=logging.INFO,
)

try:
    USE_TF = os.environ.get("USE_TF", "AUTO").upper()
    USE_TORCH = os.environ.get("USE_TORCH", "AUTO").upper()
    if USE_TORCH in ("1", "ON", "YES", "AUTO") and USE_TF not in ("1", "ON", "YES"):
        import torch
        import torchvision

        assert hasattr(torch, "__version__") and version.parse(
            torch.__version__
        ) >= version.parse("1.0.0")
        assert hasattr(torchvision, "__version__") and version.parse(
            torch.__version__
        ) >= version.parse("0.2.0")

        _torch_available = True  # pylint: disable=invalid-name
        logger.info("PyTorch version {} available.".format(torch.__version__))
        logger.info("PyTorch Vision version {} available.".format(torchvision.__version__))
    else:
        logger.info("Disabling PyTorch because USE_TF is set")
        _torch_available = False
except ImportError:
    _torch_available = False  # pylint: disable=invalid-name

try:
    USE_TF = os.environ.get("USE_TF", "AUTO").upper()
    USE_TORCH = os.environ.get("USE_TORCH", "AUTO").upper()

    if USE_TF in ("1", "ON", "YES", "AUTO") and USE_TORCH not in ("1", "ON", "YES"):
        import tensorflow as tf
        import tensorflow_io as tfio

        assert hasattr(tf, "__version__") and version.parse(
            tf.__version__
        ) >= version.parse("2.2")
        assert hasattr(tfio, "__version__") and version.parse(
            tfio.__version__
        ) >= version.parse("0.13")
        _tf_available = True  # pylint: disable=invalid-name
        logger.info("TensorFlow version {} available.".format(tf.__version__))
        logger.info("TensorFlow IO version {} available.".format(tfio.__version__))
    else:
        logger.info("Disabling Tensorflow because USE_TORCH is set")
        _tf_available = False
except (ImportError, AssertionError):
    _tf_available = False  # pylint: disable=invalid-name


def is_torch_available():
    return _torch_available


def is_tf_available():
    return _tf_available


def check_nas(endpoint):
    _url_or_path = "//" + endpoint
    u = urlsplit(_url_or_path)
    if u.path:
        return True
    if u.netloc:
        return False
    raise ValueError("This endpoint is not suitable.")



# Copyright 2020-present Tae Hwan Jung
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import uuid
import atexit
import tempfile
import tables as tb
import numpy as np
from functools import reduce
from minio import Minio

from matorage.nas import NAS
from matorage.utils import is_tf_available, is_torch_available, check_nas
from matorage.uploader import Uploader

_KB = 1024
"""The size of a Kilobyte in bytes"""

_MB = 1024 * _KB
"""The size of a Megabyte in bytes"""


class DataSaver(object):
    """

    This class must be created independently for the process. The independent process uses
    multiple threads to upload to storage and generates unique metadata information when upload is complete.
    Update the file, push the upload queue if it exceeds a certain size, close the file, and create a new file.
    After saving, you should disconnect the data saver.

    To make This procedure easier to understand, the following is written in the pseudo-code.

    .. code-block::

        per_one_batch_data_size = array_size // num_batch
        per_one_file_batch_size = max_object_size // per_one_batch_data_size
        for batch_idx in range(num_batch):
            if get_current_stored_batch_size() < per_one_file_batch_size:
                file.append(data[batch_idx])
            else:
                file_closing()
                new_file is opened
                new_file.append(data[batch_idx])
        All files are closed.

    Note:
        - Deep Learning Framework Type : All(pure python is also possible)
        - **All processes should call the constructors of this class independently.**
        - After data save is over, you must disconnect through the disconnect function.

    Args:
        config (:obj:`matorage.DataConfig`, **require**):
            A DataConfig instance object
        multipart_upload_size (:obj:`integer`, optional, defaults to `5 * 1024 * 1024`):
            size of the incompletely uploaded object.
            You can sync files faster with `multipart upload in MinIO. <https://github.com/minio/minio-py/blob/master/minio/api.py#L1795>`_
            This is because MinIO clients use multi-threading, which improves IO speed more
            efficiently regardless of Python's Global Interpreter Lock(GIL).
        num_worker_threads (:obj:`integer`, optional, defaults to 4):
            number of backend storage worker to upload or download.

        inmemory (:obj:`boolean`, optional, defaults to `False`):
            If you use this value as `True`, then you can use `HDF5_CORE driver <https://support.hdfgroup.org/HDF5/doc/TechNotes/VFL.html#TOC1>`_
            so the temporary file for uploading or downloading to backend storage,
            such as MinIO, is not stored on disk but is in the memory.
            Keep in mind that using memory is fast because it doesn't use disk IO, but it's not always good.
            If default option(False), then `HDF5_SEC2` driver will be used on posix OS(or `HDF5_WINDOWS` in Windows).

        refresh (:obj:`boolean`, optional, defaults to `False`):
            All existing data is erased and overwritten.

    Single Process example

    Examples::

        import numpy as np
        from tqdm import tqdm
        from matorage import DataConfig, DataSaver

        data_config = DataConfig(
            endpoint='127.0.0.1:9000',
            access_key='minio',
            secret_key='miniosecretkey',
            dataset_name='array_test',
            attributes=[
                ('array', 'uint8', (3, 224, 224)),
            ]
        )

        data_saver = DataSaver(config=data_config)
        row = 100
        data = np.random.rand(64, 3, 224, 224)

        for _ in tqdm(range(row)):
            data_saver({
                'array' : data
            })

        data_saver.disconnect()

    """

    def __init__(
        self,
        config,
        multipart_upload_size=5 * _MB,
        num_worker_threads=4,
        inmemory=False,
        refresh=False,
    ):

        self.config = config

        # Storage configuration
        self.multipart_upload_size = multipart_upload_size
        self.num_worker_threads = num_worker_threads

        # HDF5 configuration
        self.inmemory = inmemory

        self.filter = tb.Filters(**config.compressor)

        self._filelist = []
        self._file, self._earray = self._get_newfile()

        self._disconnected = False

        self._client = (
            Minio(
                endpoint=self.config.endpoint,
                access_key=self.config.access_key,
                secret_key=self.config.secret_key,
                secure=self.config.secure,
                region=self.config.region,
            )
            if not check_nas(self.config.endpoint)
            else NAS(self.config.endpoint)
        )
        self._check_and_create_bucket(refresh=refresh)

        self._uploader = Uploader(
            client=self._client,
            bucket=self.config.bucket_name,
            num_worker_threads=self.num_worker_threads,
            multipart_upload_size=self.multipart_upload_size,
            inmemory=self.inmemory,
        )

        atexit.register(self._exit)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disconnect()

    def _append_file(self):
        """
        upload file to key called `<bucket_name>/key`.
        appended data is `Dict[str, str]`
        **`value` is file path of `str` type**
        example:
        {
            'key' : 'value.txt',
        }

        """
        for key, filepath in self._datas.items():
            self._uploader.set_queue(
                local_file=filepath,
                remote_file=key,
            )

            self.config.set_files(key)

    def _append_numpy(self):
        """
        append numpy array in `name` node.
        appended data is `Dict[str, numpy.ndarray]` type.
        **`value` is `numpy.ndarray` type with (B, *) shape, B means batch size**
        example:
            {
                'image' : np.random.rand(16, 28, 28),
                'target' : np.random.rand(16)
            }

        """
        array_size = self._get_array_size()
        bzs = list(self._datas.values())[0].shape[0]

        per_one_batch_data_size = array_size // bzs
        per_one_file_batch_size = max(
            1, self.config.max_object_size // per_one_batch_data_size
        )

        for batch_idx in range(bzs):
            if self._get_current_stored_batch_size() < per_one_file_batch_size:
                for name, array in self._datas.items():
                    self._earray[name].append(array[batch_idx, None])
            else:
                self._file_closing()
                self._file, self._earray = self._get_newfile()
                for name, array in self._datas.items():
                    self._earray[name].append(array[batch_idx, None])

    def _check_and_create_bucket(self, refresh):
        if not self._client.bucket_exists(self.config.bucket_name):
            self._client.make_bucket(
                self.config.bucket_name, location=self.config.region
            )
        elif refresh:
            objects = self._client.list_objects(self.config.bucket_name, recursive=True)
            for obj in objects:
                self._client.remove_object(self.config.bucket_name, obj.object_name)

    def _check_attr_name(self, name):
        """
        check attribute names is exist

        """
        if name not in self._earray.keys():
            raise KeyError("attribute name {} is not exist!".format(name))

    def _check_data_filetype(self):
        """
        Check data which is file type

        """
        if not isinstance(self._datas, dict):
            raise TypeError("datas shoud be dict type.", self.__call__.__doc__)

        for key, filepath in self._datas.items():
            if not os.path.exists(filepath):
                raise FileNotFoundError("{} is not found".format(filepath))

    def _check_data_numpytype(self):
        """
        Check data which is numpy array type

        """

        if not isinstance(self._datas, dict):
            raise TypeError("datas shoud be dict type.", self.__call__.__doc__)

        bzs = 0
        for name, array in self._datas.items():
            self._check_attr_name(name=name)

            if is_tf_available() and not isinstance(array, np.ndarray):
                array = array.numpy()
            if is_torch_available() and not isinstance(array, np.ndarray):
                array = array.numpy()

            assert isinstance(array, np.ndarray), "array type is not `numpy.ndarray`"

            if bzs:
                if bzs != array.shape[0]:
                    raise ValueError("each datas array batch sizes are not same.")
            else:
                bzs = array.shape[0]

            # This resape is made into a (B, *) shape.
            # Shape is lowered to two contiguous dimensions, enabling IO operations to operate very quickly.
            # https://www.slideshare.net/HDFEOS/caching-and-buffering-in-hdf5#25
            if len(array.shape) == 1:
                # this array is ground truth
                array = array.reshape(-1, 1)

            self._datas[name] = array.reshape(
                -1, reduce(lambda x, y: x * y, array.shape[1:])
            )

    def __call__(self, datas, filetype=False):
        """

        Args:
            datas (:obj:`Dict[str, numpy.ndarray] or Dict[str, str]`, **require**):
                if filetype is false, `datas` is `Dict[str, numpy.ndarray]` type, **`value` is `numpy.ndarray` type with (B, *) shape, B means batch size**.
                else true, `datas` is `Dict[str, str]` type, **`value` is file path of `str` type**.
            filetype (:obj:`boolean`, optional):
                Indicates whether the type of data to be added to this bucket is a simple file type.

        Examples::

            data_saver = DataSaver(config=data_config)
            data_saver({
                'image' : np.random.rand(16, 28, 28),
                'target' : np.random.rand(16)
            })

        When used as shown below, filetype data is saved with a key called `<bucket_name>/raw_image`.

        Examples::

            data_saver = DataSaver(config=data_config)
            data_saver({
                'raw_image' : 'test.jpg'
            })
            print(data_config.get_filetype_list)

        """
        self._disconnected = False

        self._datas = datas

        if not filetype:
            self._check_data_numpytype()
            self._append_numpy()
        else:
            self._check_data_filetype()
            self._append_file()

    def _file_closing(self):
        _length = len(list(self._earray.values())[0])
        _last_index = self.config.get_length

        if not self.inmemory:
            self._file.close()
            self._uploader.set_queue(
                local_file=self._file.filename,
                remote_file=os.path.basename(self._filename),
            )
        else:
            self._uploader.set_queue(
                local_file=self._file.get_file_image(),
                remote_file=os.path.basename(self._filename),
            )
            self._file.close()
        # Set filename indexer
        _current_index = _last_index + _length
        self.config.set_indexer(
            {
                _current_index: {
                    "name": os.path.basename(self._filename),
                    "length": _length,
                }
            }
        )

    def _create_name(self, length=16):
        return tempfile.mktemp("{}.h5".format(uuid.uuid4().hex[:length]))

    def _exit(self):
        self._file.close()
        self._disconnected = True

    def _get_array_size(self):
        """
        Get size of all array .

        Returns:
            :obj:`datas size(bytes)`
        """
        size = 0
        for name, array in self._datas.items():
            size += array.nbytes
        return size

    def _get_current_stored_batch_size(self):
        """
        Get current file stored batch size

        Returns:
            :obj:`integer`: current stored batch size in a opened file.
        """
        return len(list(self._earray.values())[0])

    def _get_newfile(self):
        """
        Get new file inode and it's attribute

        Returns:
            :obj:`tuple(tables.File, dict)`
            second item is pytable's attribute
            {
                'name1' : tables.EArray, 'name2' : tables.EArray
            }
        """
        _driver, _driver_core_backing_store = self._set_driver()

        self._filename = self._create_name()
        self._filelist.append(self._filename)
        file = tb.open_file(
            self._filename,
            "a",
            driver=_driver,
            driver_core_backing_store=_driver_core_backing_store,
        )

        # create expandable array
        earray = {}
        for _earray in self.config.flatten_attributes:
            earray[_earray.name] = file.create_earray(
                file.root,
                _earray.name,
                _earray.type,
                shape=tuple([0]) + _earray.shape,
                filters=self.filter,
            )

        return (file, earray)

    def _get_size(self):
        if self.inmemory:
            return sys.getsizeof(self._file.get_file_image())
        else:
            return self._file.get_filesize()

    def _set_driver(self):
        """
        Setting HDF5 driver type

        Returns:
            :obj:`str` : HDF5 driver type string
        """
        if self.inmemory:
            return "H5FD_CORE", False
        else:
            if os.name == "posix":
                return "H5FD_SEC2", True
            elif os.name == "nt":
                return "H5FD_WINDOWS", True
            else:
                raise ValueError("{} OS not supported!".format(os.name))

    @property
    def get_downloaded_dataset(self):
        """
        get local paths of downloaded dataset in local storage

        Returns:
            :obj:`list`: local path of downloaded datasets
        """
        return self._filelist

    def disconnect(self):
        """
        disconnecting datasaver. close all opened files and upload to backend storage.
        Must be called after ``datasaver`` function to store data safely.

        Examples::

            data_saver = DataSaver(config=data_config)
            data_saver({
                'image' : np.random.rand(16, 28, 28),
                'target' : np.random.rand(16)
            })
            data_saver.disconnect()

        """
        self._file_closing()
        self._uploader.join_queue()

        # metadata set
        key = uuid.uuid4().hex[:16]
        _metadata_file = tempfile.mktemp(f"{key}.json")
        self.config.metadata.to_json_file(_metadata_file)
        self._client.fput_object(
            self.config.bucket_name, f"metadata/{key}.json", _metadata_file
        )
        os.remove(_metadata_file)

    @property
    def get_disconnected(self):
        return self._disconnected



# Copyright 2020-present Tae Hwan Jung
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from tables.atom import Atom, StringAtom

from matorage.serialize import Serialize


class DataAttribute(Serialize):
    """
    DataAttribute classes.

    Args:
        name (:obj:`string`, **require**):
            data attribute name .
        type (:obj:`string`, **require**):
            data attribute type. select in `string`, `bool`, `int8`, `int16`, `int32`, `int64`,
            `uint8`, `uint16`, `uint32`, `uint64`, `float32`, `float64`
        shape (:obj:`tuple`, **require**):
            data attribute shape. For example, if you specify a shape with (2, 2), you can store an array of (B, 2, 2) shapes.
        itemsize (:obj:`integer`, optional, defaults to 0):
            itemsize(bytes) for string type attribute. Must be set for string type attribute.

    Examples::

        >>> from matorage import DataAttribute
        >>> attribute = DataAttribute('array', 'uint8', (2, 2))
        >>> attribute.name
        'array'
        >>> attribute.shape
        (2, 2)
        >>> attribute.type
        UInt8Atom(shape=(), dflt=0)

    """

    type_list = [
        "string",
        "bool",
        "int8",
        "int16",
        "int32",
        "int64",
        "uint8",
        "uint16",
        "uint32",
        "uint64",
        "float32",
        "float64",
    ]

    def __init__(self, name, type, shape, itemsize=0):

        self.name = name
        self.itemsize = itemsize

        if not isinstance(type, str):
            raise TypeError("`type` is not str type")

        if type not in self.type_list:
            raise ValueError("`type` {} is unavailable type".format(type))
        elif type == "string":
            if itemsize == 0:
                raise ValueError("set `itemsize` for string type")
            self.type = StringAtom(itemsize=itemsize)
        else:
            self.type = Atom.from_type(type)

        if isinstance(shape, int):
            self.shape = (shape,)
        else:
            self.shape = shape

    def to_dict(self):
        """
        Serializes this instance to a Python dictionary.

        Returns:
            :obj:`Dict[string, any]`: Dictionary of all the attributes that make up this configuration instance

        Examples::

            >>> from matorage import DataAttribute
            >>> attribute = DataAttribute('array', 'uint8', (2, 2))
            >>> attribute.to_dict()
            {'name': 'array', 'type': 'uint8', 'shape': (2, 2)}

        """
        output = copy.deepcopy(self.__dict__)
        if hasattr(self.__class__, "type") or "type" in output:
            output["type"] = self.type.type
        if hasattr(self.__class__, "itemsize") or "itemsize" in output:
            output["itemsize"] = self.itemsize
        return output



# Copyright 2020-present Tae Hwan Jung
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import torch
import tables

from matorage.data.data import MTRData


class Dataset(torch.utils.data.Dataset, MTRData):
    """
    Dataset class for Pytorch Dataset

    This class is customized for the dataset of the PyTorch, so it is operated by the following procedure.

    1. The ``_object_file_mapper`` manages the minio object as key and the downloaded local path as value. \
    When minio object is downloaded, it is recorded in ``_object_file_maper``.
    2. We read ``_object_file_mapper`` and download only new objects that are not there.
    3. ``__getitem__`` brings numpy data in local data from data index.

    Args:
        config (:obj:`matorage.DataConfig`, **require**):
            dataset configuration
        num_worker_threads (:obj:`int`, optional, defaults to `4`):
            Number of backend storage worker to upload or download.
        clear (:obj:`boolean`, optional, defaults to `True`):
            Delete all files stored on the local storage after the program finishes.
        cache_folder_path (:obj:`str`, optional, defaults to `~/.matorage`):
            Cached folder path to check which files are downloaded complete.
        index (:obj:`boolean`, optional, defaults to `False`):
            Setting for index mode.

    Examples::

        from matorage import DataConfig
        from matorage.torch import Dataset
        from torch.utils.data import DataLoader

        data_config = DataConfig(
            endpoint='127.0.0.1:9000',
            access_key='minio',
            secret_key='miniosecretkey',
            dataset_name='array_test',
            attributes=[
                ('array', 'uint8', (3, 224, 224)),
            ]
        )

        dataset = Dataset(config=data_config, clear=True)

        # iterative mode
        for array in DataLoader(dataset):
            print(array)

        # index mode
        print(dataset[0])

    """

    def __init__(self, config, **kwargs):
        super(Dataset, self).__init__(config, **kwargs)
        self.open_files = {}

    def __len__(self):
        return self.end_indices[-1]

    def __getitem__(self, idx):
        if not self.index:
            return self._get_item_with_download(idx)
        else:
            return self._get_item_with_inmemory(idx)

    def _get_item_with_download(self, idx):
        if not self.open_files:
            self._pre_open_files()

        _objectname, _relative_index = self._find_object(idx)
        if _objectname in self._object_file_mapper:
            _open_file = self.open_files[_objectname]
            _file = _open_file["file"]
            _attr_names = _open_file["attr_names"]

            return_tensor = {}
            for _attr_name in _attr_names:
                try:
                    return_tensor[_attr_name] = self._reshape_convert_tensor(
                        numpy_array=_file.root[_attr_name][_relative_index],
                        attr_name=_attr_name,
                    )
                    if list(return_tensor[_attr_name].size()) == [1]:
                        return_tensor[_attr_name] = return_tensor[_attr_name].item()
                except:
                    raise IOError("Crash on concurrent read")

            return list(return_tensor.values())
        else:
            raise ValueError(
                "objectname({}) is not exist in {}".format(
                    _objectname, self._object_file_mapper
                )
            )

    def _exit(self):
        """
        Close all opened files and remove.

        """
        super(Dataset, self)._exit()
        for _file in list(self.open_files.values()):
            if _file["file"].isopen:
                _file["file"].close()

    def _reshape_convert_tensor(self, numpy_array, attr_name):
        """
        Reshape numpy tensor and convert from numpy to torch tensor.
        In matorage dataset save in 2D (bz, N) shape to cpu L1 cache manage dataset fast.
        Therefore, this function restores the shape for the user to use.

        Returns:
            :obj:`torch.tensor`
        """
        _shape = self.attribute[attr_name]["shape"]
        numpy_array = numpy_array.reshape(_shape)
        tensor = torch.from_numpy(numpy_array)
        return tensor

    def _pre_open_files(self):
        """
        pre-open file for each processes.
        This function call from individuallly all processes.
        Because in Pytorch Multi Processing of DataLoader use `fork` mode.

        """
        if self.index:
            raise FileNotFoundError("index mode can't not open files.")

        _driver, _driver_core_backing_store = self._set_driver()
        for _remote, _local in self._object_file_mapper.items():
            if os.path.splitext(_local)[1] != '.h5':
                continue
            _file = tables.open_file(
                _local,
                "r",
                driver=_driver,
                driver_core_backing_store=_driver_core_backing_store,
            )
            self.open_files[_remote] = {
                "file": _file,
                # "attr_names": list(_file.get_node("/")._v_children.keys()),
            }

            # Critial Bug fixed:
            # We must sort by attribute of metadata name
            self.open_files[_remote]["attr_names"] = list(self.attribute.keys())

    def _set_driver(self):
        """
        Setting HDF5 driver type

        Returns:
            :obj:`str` : HDF5 driver type string
        """

        if os.name == "posix":
            return "H5FD_SEC2", True
        elif os.name == "nt":
            return "H5FD_WINDOWS", True
        else:
            raise ValueError("{} OS not supported!".format(os.name))



# Copyright 2020-present Tae Hwan Jung
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from queue import Queue
from threading import Thread


class MTRConnector(object):
    r""" File Storage Connector class with multi-thread.
        MinIO is thread-safety, according to document.
        Although Python Global Interpreter Lock(GIL), multi thread can benefit greatly from file IO.
    """

    def __init__(self, client, bucket, num_worker_threads):
        self._client = client
        self._bucket = bucket
        self._queue = Queue()

        for i in range(num_worker_threads):
            _thread = Thread(target=self._worker)
            _thread.daemon = True
            _thread.start()

    def _worker(self):
        while True:
            _local_file, _remote_file = self._queue.get()
            self.do_job(_local_file, _remote_file)
            self._queue.task_done()

    def set_queue(self, local_file, remote_file):
        """
        Set queue of storage connector.
        `fileitem` : file bytes image or filename in local storage.
        `filename` : physical existed filename at backend storage(or will be uploaded).

        Note:
            if `config.inmemory` mode is `False`, fileitem and filename is same.

        """
        self._queue.put((local_file, remote_file,))

    def join_queue(self):
        self._queue.join()



# Original Code
# https://github.com/huggingface/transformers/blob/master/src/transformers/testing_utils.py
# See https://github.com/graykode/matorage/blob/0.1.0/NOTICE
# modified by TaeHwan Jung(@graykode)

# Copyright 2020-present Tae Hwan Jung
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from matorage.utils import _tf_available, _torch_available


def require_torch(test_case):
    """
    Decorator marking a test that requires PyTorch.

    These tests are skipped when PyTorch isn't installed.

    """
    if not _torch_available:
        test_case = unittest.skip("test requires PyTorch")(test_case)
    return test_case


def require_tf(test_case):
    """
    Decorator marking a test that requires TensorFlow.

    These tests are skipped when TensorFlow isn't installed.

    """
    if not _tf_available:
        test_case = unittest.skip("test requires TensorFlow")(test_case)
    return test_case



# Copyright 2020-present Tae Hwan Jung
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import unittest
import numpy as np
import tensorflow as tf
from tensorflow import keras

from tests.test_optimizer import OptimizerTest

from matorage.optimizer.config import OptimizerConfig
from matorage.optimizer.tensorflow.v2.manager import OptimizerManager
from matorage.testing_utils import require_tf


@require_tf
class TFOptimizerTest(OptimizerTest, unittest.TestCase):
    def create_model(self):
        model = tf.keras.models.Sequential(
            [
                keras.layers.Dense(512, activation="relu", input_shape=(784,)),
                keras.layers.Dropout(0.2),
                keras.layers.Dense(10),
            ]
        )
        model.compile(
            optimizer="adam",
            loss=tf.losses.SparseCategoricalCrossentropy(from_logits=True),
            metrics=["accuracy"],
        )
        return model

    def test_optimizer_saver(self, optimizer_config=None, save_to_json_file=False):

        (train_images, train_labels), _ = tf.keras.datasets.mnist.load_data()
        train_labels = train_labels[:1000]
        train_images = train_images[:1000].reshape(-1, 28 * 28) / 255.0

        self.model = self.create_model()
        self.model.fit(train_images, train_labels, epochs=1)

        if optimizer_config is None:
            self.optimizer_config = OptimizerConfig(
                **self.storage_config,
                optimizer_name="test_optimizer_saver",
                additional={"framework": "tensorflow"}
            )
        else:
            self.optimizer_config = optimizer_config

        if save_to_json_file:
            self.optimizer_config_file = "model_config_file.json"
            self.optimizer_config.to_json_file(self.optimizer_config_file)

        self.optimizer_manager = OptimizerManager(config=self.optimizer_config)

        self.optimizer_manager.save(self.model.optimizer)

    def test_tfmodel_saver_from_json_file(self):

        self.test_optimizer_saver(save_to_json_file=True)

        self.optimizer_config = None
        self.optimizer_manager = None

        self.optimizer_config = OptimizerConfig.from_json_file(
            self.optimizer_config_file
        )

        self.optimizer_manager = OptimizerManager(config=self.optimizer_config)

        self.optimizer_manager.save(self.model.optimizer)

    def test_optimizer_loader(self):

        (train_images, train_labels), _ = tf.keras.datasets.mnist.load_data()
        train_labels = train_labels[:1000]
        train_images = train_images[:1000].reshape(-1, 28 * 28) / 255.0

        self.model = self.create_model()
        self.model.fit(train_images, train_labels, epochs=1)
        before_optim_weight = copy.deepcopy(self.model.optimizer.get_weights())

        self.test_optimizer_saver()

        self.optimizer_manager.load(self.model.optimizer, step=32)
        after_optim_weight = self.model.optimizer.get_weights()

        assert len(before_optim_weight) == len(after_optim_weight)
        for i in range(1, len(after_optim_weight)):
            assert not np.array_equal(after_optim_weight, before_optim_weight)

    def test_optimizer_saver_nas(self):

        (train_images, train_labels), _ = tf.keras.datasets.mnist.load_data()
        train_labels = train_labels[:1000]
        train_images = train_images[:1000].reshape(-1, 28 * 28) / 255.0

        self.model = self.create_model()
        self.model.fit(train_images, train_labels, epochs=1)

        self.optimizer_config = OptimizerConfig(
            **self.nas_config,
            optimizer_name="test_optimizer_saver_nas",
            additional={"framework": "tensorflow"}
        )

        self.optimizer_manager = OptimizerManager(config=self.optimizer_config)

        self.optimizer_manager.save(self.model.optimizer)

    def test_optimizer_loader_nas(self):

        (train_images, train_labels), _ = tf.keras.datasets.mnist.load_data()
        train_labels = train_labels[:1000]
        train_images = train_images[:1000].reshape(-1, 28 * 28) / 255.0

        self.model = self.create_model()
        self.model.fit(train_images, train_labels, epochs=1)
        before_optim_weight = copy.deepcopy(self.model.optimizer.get_weights())

        self.test_optimizer_saver_nas()

        self.optimizer_manager.load(self.model.optimizer, step=32)
        after_optim_weight = self.model.optimizer.get_weights()

        assert len(before_optim_weight) == len(after_optim_weight)
        for i in range(1, len(after_optim_weight)):
            assert not np.array_equal(after_optim_weight, before_optim_weight)


def suite():
    return unittest.TestSuite(unittest.makeSuite(TFOptimizerTest))


if __name__ == "__main__":
    unittest.main(defaultTest="suite")
